"""
A collection of convenience functions for the preparation and conversion of [`NetworkX`](https://networkx.github.io/)
graphs to and from `cityseer` data structures. Note that the `cityseer` network data structures can be created and
manipulated directly, if so desired.
"""

import json
import logging
from typing import Union, Tuple, Optional

import networkx as nx
import numpy as np
import utm
from numba import types
from numba.typed import Dict
from shapely import geometry, ops, strtree, coords
from tqdm.auto import tqdm

from cityseer.algos import checks

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def nX_simple_geoms(networkX_multigraph: nx.MultiGraph) -> nx.MultiGraph:
    """
    Generates straight-line geometries for each edge based on the the `x` and `y` coordinates of the adjacent nodes.
    The edge geometry will be stored to the edge `geom` attribute.

    Parameters
    ----------
    networkX_multigraph
        A `networkX` `MultiGraph` with `x` and `y` node attributes.

    Returns
    -------
    nx.MultiGraph
        A `networkX` `MultiGraph` with `shapely` [`Linestring`](https://shapely.readthedocs.io/en/latest/manual.html#linestrings)
        geometries assigned to the edge `geom` attributes.
    """
    if not isinstance(networkX_multigraph, nx.MultiGraph):
        raise TypeError('This method requires an undirected networkX MultiGraph.')
    logger.info('Generating simple (straight) edge geometries.')
    g_multi_copy = networkX_multigraph.copy()
    # unpack coordinates and build simple edge geoms
    for s, e, k in tqdm(g_multi_copy.edges(keys=True), disable=checks.quiet_mode):
        # start x coordinate
        if 'x' not in g_multi_copy.nodes[s]:
            raise KeyError(f'Encountered node missing "x" coordinate attribute at node {s}.')
        s_x = g_multi_copy.nodes[s]['x']
        # start y coordinate
        if 'y' not in g_multi_copy.nodes[s]:
            raise KeyError(f'Encountered node missing "y" coordinate attribute at node {s}.')
        s_y = g_multi_copy.nodes[s]['y']
        # end x coordinate
        if 'x' not in g_multi_copy.nodes[e]:
            raise KeyError(f'Encountered node missing "x" coordinate attribute at node {e}.')
        e_x = g_multi_copy.nodes[e]['x']
        # end y coordinate
        if 'y' not in g_multi_copy.nodes[e]:
            raise KeyError(f'Encountered node missing "y" coordinate attribute at node {e}.')
        e_y = g_multi_copy.nodes[e]['y']
        g_multi_copy[s][e][k]['geom'] = geometry.LineString([[s_x, s_y], [e_x, e_y]])

    return g_multi_copy


def _add_node(networkX_multigraph: nx.MultiGraph,
              node_names: Union[list, tuple],
              x: float,
              y: float,
              live: Optional[bool] = None) -> Optional[str]:
    """
    Adds a node to a networkX `MultiGraph`. Assembles a new name from source node names. Checks for duplicates.
    """
    # suggest a name based on the given names
    if len(node_names) == 1:
        new_nd_name = str(node_names[0])
    # if concatenating existing nodes, suggest a name based on a combination of existing names
    else:
        names = []
        for name in node_names:
            name = str(name)
            if len(name) > 10:
                name = f'{name[:5]}|{name[-5:]}'
            names.append(name)
        new_nd_name = '±'.join(names)
    # first check whether the node already exists
    append = 2
    target_name = new_nd_name
    dupe = False
    while True:
        if f'{new_nd_name}' in networkX_multigraph:
            dupe = True
            # if the coordinates also match, then it is probable that the same node is being re-added...
            nd = networkX_multigraph.nodes[f'{new_nd_name}']
            if nd['x'] == x and nd['y'] == y:
                logger.debug(f'Proposed new node {new_nd_name} would overlay a node that already exists '
                             f'at the same coordinates. Skipping.')
                return None
            # otherwise, warn and bump the appended node number
            new_nd_name = f'{target_name}§v{append}'
            append += 1
        else:
            if dupe:
                logger.debug(f'A node of the same name already exists in the graph, '
                             f'adding this node as {new_nd_name} instead.')
            break
    # add
    attributes = {'x': x, 'y': y}
    if live is not None:
        attributes['live'] = live
    networkX_multigraph.add_node(new_nd_name, **attributes)
    return new_nd_name


def nX_from_osm(osm_json: str) -> nx.MultiGraph:
    """
    Generates a `NetworkX` `MultiGraph` from [Open Street Map](https://www.openstreetmap.org) data.

    Parameters
    -----------
    osm_json
        A `json` string response from the [OSM overpass API](https://wiki.openstreetmap.org/wiki/Overpass_API),
        consisting of `nodes` and `ways`.

    Returns
    -------
    nx.MultiGraph
        A `NetworkX` `MultiGraph` with `x` and `y` attributes in [WGS84](https://epsg.io/4326) `lng`, `lat` geographic
        coordinates.

    """
    osm_network_data = json.loads(osm_json)
    G = nx.MultiGraph()
    for e in osm_network_data['elements']:
        if e['type'] == 'node':
            G.add_node(e['id'], x=e['lon'], y=e['lat'])
    for e in osm_network_data['elements']:
        if e['type'] == 'way':
            count = len(e['nodes'])
            for idx in range(count - 1):
                G.add_edge(e['nodes'][idx], e['nodes'][idx + 1])

    return G


def nX_wgs_to_utm(networkX_multigraph: nx.MultiGraph, force_zone_number: int=None) -> nx.MultiGraph:
    """
    Converts `x` and `y` node attributes from [WGS84](https://epsg.io/4326) `lng`, `lat` geographic coordinates to the
    local UTM projected coordinate system. If edge `geom` attributes are found, the associated `LineString` geometries
    will also be converted. The UTM zone derived from the first processed node will be used for the conversion of all
    other nodes and geometries contained in the graph. This ensures consistent behaviour in cases where a graph spans
    a UTM boundary.

    Parameters
    ----------
    networkX_multigraph
        A `networkX` `MultiGraph` with `x` and `y` node attributes in the WGS84 coordinate system. Optional `geom` edge
        attributes containing `LineString` geoms to be converted.
    force_zone_number
        An optional UTM zone number for coercing all conversions to an explicit UTM zone. Use with caution: mismatched
        UTM zones may introduce substantial distortions in the results. Defaults to None.

    Returns
    -------
    nx.MultiGraph
        A `networkX` `MultiGraph` with `x` and `y` node attributes converted to the local UTM coordinate system. If edge
         `geom` attributes are present, these will also be converted.
    """
    if not isinstance(networkX_multigraph, nx.MultiGraph):
        raise TypeError('This method requires an undirected networkX MultiGraph.')
    logger.info('Converting networkX graph from WGS to UTM.')
    g_multi_copy = networkX_multigraph.copy()
    zone_number = None
    if force_zone_number is not None:
        zone_number = force_zone_number
    logger.info('Processing node x, y coordinates.')
    for n, d in tqdm(g_multi_copy.nodes(data=True), disable=checks.quiet_mode):
        # x coordinate
        if 'x' not in d:
            raise KeyError(f'Encountered node missing "x" coordinate attribute at node {n}.')
        lng = d['x']
        # y coordinate
        if 'y' not in d:
            raise KeyError(f'Encountered node missing "y" coordinate attribute at node {n}.')
        lat = d['y']
        # check for unintentional use of conversion
        if abs(lng) > 180 or abs(lat) > 90:
            raise ValueError('x, y coordinates exceed WGS bounds. Please check your coordinate system.')
        # to avoid issues across UTM boundaries, use the first point to set (and subsequently force) the UTM zone
        if zone_number is None:
            zone_number = utm.from_latlon(lat, lng)[2]  # zone number is position 2
        # be cognisant of parameter and return order
        # returns in easting, northing order
        easting, northing = utm.from_latlon(lat, lng, force_zone_number=zone_number)[:2]
        # write back to graph
        g_multi_copy.nodes[n]['x'] = easting
        g_multi_copy.nodes[n]['y'] = northing
    # if line geom property provided, then convert as well
    logger.info('Processing edge geom coordinates, if present.')
    for s, e, k, d in tqdm(g_multi_copy.edges(data=True, keys=True), disable=checks.quiet_mode):
        # check if geom present - optional step
        if 'geom' in d:
            line_geom = d['geom']
            if line_geom.type != 'LineString':
                raise TypeError(f'Expecting LineString geometry but found {line_geom.type} geometry.')
            # be cognisant of parameter and return order
            # returns in easting, northing order
            utm_coords = [utm.from_latlon(lat, lng, force_zone_number=zone_number)[:2] for lng, lat in line_geom.coords]
            # write back to edge
            g_multi_copy[s][e][k]['geom'] = geometry.LineString(utm_coords)

    return g_multi_copy


def nX_remove_dangling_nodes(networkX_multigraph: nx.MultiGraph,
                             despine: float = None,
                             remove_disconnected: bool = True) -> nx.MultiGraph:
    """
    Optionally removes short dead-ends or disconnected graph components, which may be prevalent on poor quality network
    datasets.

    Parameters
    ----------
    networkX_multigraph
        A `networkX` `MultiGraph` in a projected coordinate system, containing `x` and `y` node attributes, and `geom`
        edge attributes containing `LineString` geoms.
    despine
        The maximum cutoff distance for removal of dead-ends. Use `0` where no despining should occur. Defaults to None.
    remove_disconnected
        Whether to remove disconnected components. If set to `True`, only the largest connected component will be
        returned. Defaults to True.

    Returns
    -------
    nx.MultiGraph
        A `networkX` `MultiGraph` with disconnected components optionally removed, and dead-ends removed where less than
         the `despine` parameter distance.
    """
    logger.info(f'Removing dangling nodes.')
    g_multi_copy = networkX_multigraph.copy()
    if remove_disconnected:
        # finds connected components - this behaviour changed with networkx v2.4
        connected_components = list(nx.algorithms.components.connected_components(g_multi_copy))
        # sort by largest component
        g_nodes = sorted(connected_components, key=len, reverse=True)[0]
        # make a copy of the graph using the largest component
        g_multi_copy = nx.MultiGraph(g_multi_copy.subgraph(g_nodes))
    if despine:
        remove_nodes = []
        for n, d in tqdm(g_multi_copy.nodes(data=True), disable=checks.quiet_mode):
            if nx.degree(g_multi_copy, n) == 1:
                # only a single neighbour, so index-in directly and update at key = 0
                nb = list(nx.neighbors(g_multi_copy, n))[0]
                if g_multi_copy[n][nb][0]['geom'].length <= despine:
                    remove_nodes.append(n)
        g_multi_copy.remove_nodes_from(remove_nodes)

    return g_multi_copy


def _snap_linestring_idx(linestring_coords: Union[list, tuple, np.ndarray, coords.CoordinateSequence],
                         idx: int,
                         xy: Tuple[float, float]) -> list:
    """
    Snaps a LineString's coordinate at the specified index to the provided xy coordinate.
    """
    # check types
    if not isinstance(linestring_coords, (list, tuple, np.ndarray, coords.CoordinateSequence)):
        raise ValueError('Expecting a list, tuple, numpy array, or shapely LineString coordinate sequence.')
    linestring_coords = list(linestring_coords)
    # check that the index is either 0 or 1
    if idx not in [0, -1]:
        raise ValueError('Expecting either a start index of "0" or an end index of "-1"')
    # handle 3D
    coord = list(linestring_coords[idx])  # tuples don't support indexed assignment
    coord[:2] = xy
    linestring_coords[idx] = tuple(coord)

    return linestring_coords


def _snap_linestring_startpoint(linestring_coords: Union[list, tuple, np.ndarray, coords.CoordinateSequence],
                                xy: Tuple[float, float]) -> list:
    """
    Snaps a LineString's start-point coordinate to a specified xy coordinate.
    """
    return _snap_linestring_idx(linestring_coords, 0, xy)


def _snap_linestring_endpoint(linestring_coords: Union[list, tuple, np.ndarray, coords.CoordinateSequence],
                              xy: Tuple[float, float]) -> list:
    """
    Snaps a LineString's end-point coordinate to a specified xy coordinate.
    """
    return _snap_linestring_idx(linestring_coords, -1, xy)


def _align_linestring_coords(linestring_coords: Union[list, tuple, np.ndarray, coords.CoordinateSequence],
                             xy: Tuple[float, float],
                             reverse: bool = False,
                             tolerance=checks.tolerance) -> list:
    """
    Aligns a LineString's coordinate order to either start or end at the xy coordinate within a given tolerance.
    If reverse=False the coordinate order will be aligned to start from the given xy coordinate.
    If reverse=True the coordinate order will be aligned to end at the given xy coordinate.
    """
    # check types
    if not isinstance(linestring_coords, (list, tuple, np.ndarray, coords.CoordinateSequence)):
        raise ValueError('Expecting a list, tuple, numpy array, or shapely LineString coordinate sequence.')
    linestring_coords = list(linestring_coords)
    # the target indices depend on whether reversed or not
    if not reverse:
        xy_idx = 0
        opposite_idx = -1
    else:
        xy_idx = -1
        opposite_idx = 0
    # flip if necessary
    if np.allclose(xy, linestring_coords[opposite_idx][:2], atol=tolerance, rtol=0):
        return linestring_coords[::-1]
    # if still not aligning, then there is an issue
    elif not np.allclose(xy, linestring_coords[xy_idx][:2], atol=tolerance, rtol=0):
        raise ValueError(f'Unable to align the LineString to starting point {xy} given the tolerance of {tolerance}.')
    # otherwise no flipping is required and the coordinates can simply be returned
    else:
        return linestring_coords


def _weld_linestring_coords(linestring_coords_a: Union[list, tuple, np.ndarray, coords.CoordinateSequence],
                            linestring_coords_b: Union[list, tuple, np.ndarray, coords.CoordinateSequence],
                            force_xy: Tuple[float, float] = None,
                            tolerance=checks.tolerance) -> list:
    """
    Takes two geometries, finds a matching start / end point combination and merges the coordinates accordingly.
    If the optional force_xy is provided then the weld will be performed at the xy end of the LineStrings.
    The force_xy parameter is useful for looping geometries or overlapping geometries where it can happen that
    welding works from either of the two ends, thus potentially mis-aligning the start point unless explicit.
    """
    # check types
    for lc in [linestring_coords_a, linestring_coords_b]:
        if not isinstance(lc, (list, tuple, np.ndarray, coords.CoordinateSequence)):
            raise ValueError('Expecting a list, tuple, numpy array, or shapely LineString coordinate sequence.')
    linestring_coords_a = list(linestring_coords_a)
    linestring_coords_b = list(linestring_coords_b)
    # if both lists are empty, raise
    if len(linestring_coords_a) == 0 and len(linestring_coords_b) == 0:
        raise ValueError('Neither of the provided linestring coordinate lists contain any coordinates.')
    # if one of the lists is empty, return only the other
    elif not len(linestring_coords_b):
        return linestring_coords_a
    elif not len(linestring_coords_a):
        return linestring_coords_b
    # match the directionality of the linestrings
    # if override_xy is provided, then make sure that the sides with the specified xy are merged
    # this is useful for looping components or overlapping components
    # i.e. where both the start and end points match an endpoint on the opposite line
    # in this case it is necessary to know which is the inner side of the weld and which is the outer endpoint
    if force_xy:
        if not np.allclose(linestring_coords_a[-1][:2], force_xy, atol=tolerance, rtol=0):
            coords_a = _align_linestring_coords(linestring_coords_a, force_xy, reverse=True)
        else:
            coords_a = linestring_coords_a
        if not np.allclose(linestring_coords_b[0][:2], force_xy, atol=tolerance, rtol=0):
            coords_b = _align_linestring_coords(linestring_coords_b, force_xy, reverse=False)
        else:
            coords_b = linestring_coords_b
    # case A: the linestring_b has to be flipped to start from x, y
    elif np.allclose(linestring_coords_a[-1][:2], linestring_coords_b[-1][:2], atol=tolerance, rtol=0):
        anchor_xy = linestring_coords_a[-1][:2]
        coords_a = linestring_coords_a
        coords_b = _align_linestring_coords(linestring_coords_b, anchor_xy)
    # case B: linestring_a has to be flipped to end at x, y
    elif np.allclose(linestring_coords_a[0][:2], linestring_coords_b[0][:2], atol=tolerance, rtol=0):
        anchor_xy = linestring_coords_a[0][:2]
        coords_a = _align_linestring_coords(linestring_coords_a, anchor_xy)
        coords_b = linestring_coords_b
    # case C: merge in the b -> a order (saves flipping both)
    elif np.allclose(linestring_coords_a[0][:2], linestring_coords_b[-1][:2], atol=tolerance, rtol=0):
        coords_a = linestring_coords_b
        coords_b = linestring_coords_a
    # case D: no further alignment is necessary
    else:
        coords_a = linestring_coords_a
        coords_b = linestring_coords_b
    # double check weld
    if not np.allclose(coords_a[-1][:2], coords_b[0][:2], atol=tolerance, rtol=0):
        raise ValueError(f'Unable to weld LineString geometries with the given tolerance of {tolerance}.')
    # drop the duplicate interleaving coordinate
    return coords_a[:-1] + coords_b


def nX_remove_filler_nodes(networkX_multigraph: nx.MultiGraph) -> nx.MultiGraph:
    """
    Removes nodes of degree=2: such nodes represent no route-choices other than traversal to the next edge.
    The edges on either side of the deleted nodes will be removed and replaced with a new spliced edge.

    :::tip Note
    Filler nodes may be prevalent in poor quality datasets, or in situations where curved roadways have been represented
    through the addition of nodes to describe arced geometries. `cityseer` uses `shapely` [`Linestrings`](https://shapely.readthedocs.io/en/latest/manual.html#linestrings)
    to describe arbitrary road geometries without the need for filler nodes. Filler nodes can therefore be removed, thus
    reducing side-effects when computing network centralities, which arise as a function of varied node intensities.
    :::

    Parameters
    ----------
    networkX_multigraph
        A `networkX` `MultiGraph` in a projected coordinate system, containing `x` and `y` node attributes, and `geom`
        edge attributes containing `LineString` geoms.

    Returns
    -------
    nx.MultiGraph
        A `networkX` `MultiGraph` with nodes of degree=2 removed. Adjacent edges will be combined into a unified new
        edge with associated `geom` attributes spliced together.
    """
    if not isinstance(networkX_multigraph, nx.MultiGraph):
        raise TypeError('This method requires an undirected networkX MultiGraph.')
    logger.info(f'Removing filler nodes.')
    g_multi_copy = networkX_multigraph.copy()
    removed_nodes = set()
    # iterates the original graph, but changes are written to the copied version (to avoid in-place snafus)
    for n in tqdm(networkX_multigraph.nodes(), disable=checks.quiet_mode):
        # some nodes will already have been removed
        if n in removed_nodes:
            continue
        # proceed if a "simple" node is discovered, i.e. degree = 2
        if nx.degree(networkX_multigraph, n) == 2:
            # pick the first neighbour and follow the chain until a non-simple node is encountered
            # this will become the starting point of the chain of simple nodes to be consolidated
            nbs = list(nx.neighbors(networkX_multigraph, n))
            # catch the edge case where the a single dead-end node has two out-edges to a single neighbour
            if len(nbs) == 1:
                continue
            # if only one neighbour, then ignore (e.g. dead-end with two edges linking back to another node)
            if nbs == 1:
                continue
            # otherwise randomly select one side and find a non-simple node as a starting point.
            else:
                nb = nbs[0]
            # anchor_nd should be the first node of the chain of nodes to be merged, and should be a non-simple node
            anchor_nd = None
            # next_link_nd should be a direct neighbour of anchor_nd and must be a simple node
            next_link_nd = n
            # find the non-simple start node
            while anchor_nd is None:
                # follow the chain of neighbours and break once a non-simple node is found
                # catch disconnected looping components by checking for re-encountering start-node
                if nx.degree(networkX_multigraph, nb) != 2 or nb == n:
                    anchor_nd = nb
                    break
                # probe neighbours in one-direction only - i.e. don't backtrack
                nb_a, nb_b = list(nx.neighbors(networkX_multigraph, nb))
                if nb_a == next_link_nd:
                    next_link_nd = nb
                    nb = nb_b
                else:
                    next_link_nd = nb
                    nb = nb_a
            # from anchor_nd, proceed along the chain in the next_link_nd direction
            # accumulate and weld geometries along the way
            # break once finding another non-simple node
            trailing_nd = anchor_nd
            end_nd = None
            drop_nodes = []
            agg_geom = []
            while end_nd is None:
                # aggregate the geom
                try:
                    # there is ordinarily a single edge from trailing to next
                    # however, there is an edge case where next is a dead-end with two edges linking back to trailing
                    # (i.e. where one of those edges is longer than the maximum length discrepancy for merging edges)
                    # in either case, use the first geom
                    geom = networkX_multigraph[trailing_nd][next_link_nd][0]['geom']
                except KeyError:
                    raise KeyError(f'Missing "geom" attribute for edge {trailing_nd}-{next_link_nd}')
                if geom.type != 'LineString':
                    raise TypeError(f'Expecting LineString geometry but found {geom.type} geometry.')
                # welds can be done automatically, but there are edge cases, e.g.:
                # looped roadways or overlapping edges such as stairways don't know which sides of two segments to join
                # i.e. in these cases the edges can sometimes be matched from one of two possible configurations
                # since the xy join is known for all cases it is used here regardless
                override_xy = (networkX_multigraph.nodes[trailing_nd]['x'], networkX_multigraph.nodes[trailing_nd]['y'])
                # weld
                agg_geom = _weld_linestring_coords(agg_geom, geom.coords, force_xy=override_xy)
                # if the next node has a degree other than 2, then break
                # for circular components, break if the next node matches the start node
                if nx.degree(networkX_multigraph, next_link_nd) != 2 or next_link_nd == anchor_nd:
                    end_nd = next_link_nd
                # otherwise, follow the chain
                else:
                    # add next_link_nd to drop list
                    drop_nodes.append(next_link_nd)
                    # get the next set of neighbours
                    # in the above-mentioned edge-case, a single dead-end node with two edges back to a start node
                    # will only have one neighbour
                    new_nbs = list(nx.neighbors(networkX_multigraph, next_link_nd))
                    if len(new_nbs) == 1:
                        trailing_nd = next_link_nd
                        next_link_nd = new_nbs[0]
                    # but in almost all cases there will be two neighbours, one of which will be the previous node
                    else:
                        nb_a, nb_b = list(nx.neighbors(networkX_multigraph, next_link_nd))
                        # proceed to the new_next node
                        if nb_a == trailing_nd:
                            trailing_nd = next_link_nd
                            next_link_nd = nb_b
                        else:
                            trailing_nd = next_link_nd
                            next_link_nd = nb_a
            # double-check that the geom's endpoints match within tolerance
            # then snap to remove any potential side-effects from minor tolerance issues
            s_xy = (networkX_multigraph.nodes[anchor_nd]['x'], networkX_multigraph.nodes[anchor_nd]['y'])
            if not np.allclose(agg_geom[0], s_xy, atol=checks.tolerance, rtol=0):
                raise ValueError('New Linestring geometry does not match starting node coordinates.')
            else:
                agg_geom = _snap_linestring_startpoint(agg_geom, s_xy)
            e_xy = (networkX_multigraph.nodes[end_nd]['x'], networkX_multigraph.nodes[end_nd]['y'])
            if not np.allclose(agg_geom[-1], e_xy, atol=checks.tolerance, rtol=0):
                raise ValueError('New Linestring geometry does not match ending node coordinates.')
            else:
                agg_geom = _snap_linestring_endpoint(agg_geom, e_xy)
            # create a new linestring
            new_geom = geometry.LineString(agg_geom)
            if new_geom.type != 'LineString':
                raise TypeError(
                    f'Found {new_geom.type} geometry instead of "LineString" for new geom {new_geom.wkt}.'
                    f'Check that the adjacent LineStrings in the vicinity of {n} are not corrupted.')
            # add a new edge from anchor_nd to end_nd
            g_multi_copy.add_edge(anchor_nd, end_nd, geom=new_geom)
            # drop the removed nodes, which will also implicitly drop the related edges
            g_multi_copy.remove_nodes_from(drop_nodes)
            removed_nodes.update(drop_nodes)

    return g_multi_copy


def _squash_adjacent(networkX_multigraph: nx.MultiGraph,
                     node_group: Union[set, list, tuple],
                     cent_min_degree: Optional[int] = None,
                     cent_min_len_factor: Optional[float] = None) -> nx.MultiGraph:
    """
    Squashes nodes from the node group down to a new node. The new node can either be based on:
    - The centroid of all nodes;
    - Else, all nodes of degree greater or equal to cent_min_degree;
    - Else, all nodes with aggregate adjacent edge lengths greater than cent_min_len_factor as a factor of the node with
      the greatest overall aggregate lengths. Edges are adjusted from the old nodes to the new combined node.
    """
    if not isinstance(networkX_multigraph, nx.MultiGraph):
        raise TypeError('This method requires an undirected networkX MultiGraph (for multiple edges).')
    if cent_min_degree is not None and cent_min_degree < 1:
        raise ValueError('merge_node_min_degree should be a positive integer.')
    if cent_min_len_factor is not None and not 1 >= cent_min_len_factor >= 0:
        raise ValueError('cent_min_len_factor should be a decimal between 0 and 1.')
    # remove any node uids no longer in the graph
    node_group = [n for n in node_group if n in networkX_multigraph]
    # filter out nodes if using cent_min_degree or cent_min_len_factor
    filtered_nodes = []
    if cent_min_degree is not None:
        for n_uid in node_group:
            if nx.degree(networkX_multigraph, n_uid) >= cent_min_degree:
                filtered_nodes.append(n_uid)
    # else if merging on a longest adjacent edges basis
    if cent_min_len_factor is not None:
        # if nodes are pre-filtered by edge degrees, then use the filtered nodes as a starting point
        if filtered_nodes:
            node_pool = filtered_nodes.copy()
            filtered_nodes = []  # reset
        # else use the full original node group
        else:
            node_pool = node_group
        agg_lens = []
        for n_uid in node_pool:
            agg_len = 0
            # iterate each node's neighbours, aggregating neighbouring edge lengths along the way
            for nb_uid in nx.neighbors(networkX_multigraph, n_uid):
                for nb_edge in networkX_multigraph[n_uid][nb_uid].values():
                    agg_len += nb_edge['geom'].length
            agg_lens.append(agg_len)
        # find the longest
        max_len = max(agg_lens)
        # select all nodes with an agg_len within a small tolerance of longest
        for n_uid, agg_len in zip(node_pool, agg_lens):
            if agg_len >= max_len * cent_min_len_factor:
                filtered_nodes.append(n_uid)
    # otherwise, derive the centroid from all nodes
    # this is also a fallback if no nodes selected via minimum degree basis
    if not filtered_nodes:
        filtered_nodes = node_group
    # prepare the names and geoms for all points used for the new centroid
    node_geoms = []
    coords_set = set()
    for n_uid in filtered_nodes:
        x = networkX_multigraph.nodes[n_uid]['x']
        y = networkX_multigraph.nodes[n_uid]['y']
        # in rare cases opposing geom splitting can cause overlaying nodes
        # these can swing the gravity of multipoint centroids so screen these out
        xy_key = f'{round(x)}-{round(y)}'
        if xy_key in coords_set:
            continue
        else:
            coords_set.add(xy_key)
            node_geoms.append(geometry.Point(x, y))
    # set the new centroid from the centroid of the node group's Multipoint:
    c = geometry.MultiPoint(node_geoms).centroid
    # now that the centroid is known, go ahead and merge the _node_group
    # add the new node
    new_nd_name = _add_node(networkX_multigraph, node_group, x=c.x, y=c.y)
    if new_nd_name is None:
        raise ValueError(f'Attempted to add a duplicate node for node_group {node_group}.')
    # iterate the nodes to be removed and connect their existing edge geometries to the new centroid
    for uid in node_group:
        # iterate the node's existing neighbours
        for nb_uid in nx.neighbors(networkX_multigraph, uid):
            # if a neighbour is also going to be dropped, then no need to create new between edges
            # an exception exists when a geom is looped, in which case the neighbour is also the current node
            if nb_uid in node_group and nb_uid != uid:
                continue
            # MultiGraph - so iter edges
            for edge in networkX_multigraph[uid][nb_uid].values():
                if 'geom' not in edge:
                    raise KeyError(f'Missing "geom" attribute for edge {uid}-{nb_uid}')
                line_geom = edge['geom']
                if line_geom.type != 'LineString':
                    raise TypeError(
                        f'Expecting LineString geometry but found {line_geom.type} geometry for edge {uid}-{nb_uid}.')
                # orient the LineString so that the starting point matches the node's xy
                s_xy = (networkX_multigraph.nodes[uid]['x'], networkX_multigraph.nodes[uid]['y'])
                line_coords = _align_linestring_coords(line_geom.coords, s_xy)
                # update geom starting point to new parent node's coordinates
                line_coords = _snap_linestring_startpoint(line_coords, (c.x, c.y))
                # if self-loop, then the end also needs updating
                if uid == nb_uid:
                    line_coords = _snap_linestring_endpoint(line_coords, (c.x, c.y))
                    target_uid = new_nd_name
                else:
                    target_uid = nb_uid
                # build the new geom
                new_edge_geom = geometry.LineString(line_coords)
                # check that a duplicate is not being added
                dupe = False
                if networkX_multigraph.has_edge(new_nd_name, target_uid):
                    # only add parallel edges if substantially different from any existing edges
                    n_edges = networkX_multigraph.number_of_edges(new_nd_name, target_uid)
                    for k in range(n_edges):
                        existing_edge_geom = networkX_multigraph[new_nd_name][target_uid][k]["geom"]
                        # don't add if the edges have the same number of coords and the coords are similar
                        # 5m x and y tolerance across all coordinates
                        if len(new_edge_geom.coords) == len(existing_edge_geom.coords) and \
                                np.allclose(new_edge_geom.coords, existing_edge_geom.coords, atol=5, rtol=0):
                            dupe = True
                            logger.debug(
                                f'Not adding edge {new_nd_name} to {target_uid}: a similar edge already exists. '
                                f'Length: {new_edge_geom.length} vs. {existing_edge_geom.length}. '
                                f'Num coords: {len(new_edge_geom.coords)} vs. {len(existing_edge_geom.coords)}.')
                if not dupe:
                    # add the new edge
                    networkX_multigraph.add_edge(new_nd_name, target_uid, geom=new_edge_geom)
        # drop the node, this will also implicitly drop the old edges
        networkX_multigraph.remove_node(uid)

    return networkX_multigraph


def _merge_parallel_edges(networkX_multigraph: nx.MultiGraph,
                          merge_edges_by_midline: bool,
                          multi_edge_len_factor: float,
                          multi_edge_min_len: float) -> nx.MultiGraph:
    """
    Checks a MultiGraph for duplicate edges, which are then consolidated.
    If merge_edges_by_midline is False, then the shortest of the edges is used and the others are simply dropped.
    If merge_edges_by_midline is True, then the duplicates are replaced with a new edge following the merged centreline.
    In cases where one line is significantly longer than another (e.g. a crescent streets),
    then the longer edge is retained as separate if exceeding the multi_edge_len_factor as a factor of the shortest
    length but with the exception that (longer) edges still shorter than multi_edge_min_len are removed regardless.
    """
    if not isinstance(networkX_multigraph, nx.MultiGraph):
        raise TypeError('This method requires an undirected networkX MultiGraph (for multiple edges).')
    if multi_edge_len_factor <= 1:
        raise TypeError('multi_edge_len_factor should be a factor greater than 1. ')
    if multi_edge_len_factor < 1.25:
        logger.warning('Merging by midline and setting multi_edge_len_factor too low (e.g. lower than 1.25) may '
                       'result in an undesirable number of relatively similar parallel edges.')
    # don't use copy() - add nodes only
    deduped_graph = nx.MultiGraph()
    deduped_graph.add_nodes_from(networkX_multigraph.nodes(data=True))
    # iter the edges
    for s, e, d in tqdm(networkX_multigraph.edges(data=True), disable=checks.quiet_mode):
        # if only one edge is associated with this node pair, then add
        if networkX_multigraph.number_of_edges(s, e) == 1:
            deduped_graph.add_edge(s, e, **d)
        # otherwise, add if not already added from another (parallel) edge
        elif not deduped_graph.has_edge(s, e):
            # there are normally two edges, but sometimes three or possibly more
            edges = networkX_multigraph[s][e].values()
            # find the shortest of the geoms
            edge_geoms = [edge['geom'] for edge in edges]
            edge_lens = [geom.length for geom in edge_geoms]
            shortest_idx = edge_lens.index(min(edge_lens))
            shortest_len = edge_lens.pop(shortest_idx)
            shortest_geom = edge_geoms.pop(shortest_idx)
            longer_geoms = []
            for edge_len, edge_geom in zip(edge_lens, edge_geoms):
                # retain distinct edges where they are substantially longer than the shortest geom
                if edge_len > shortest_len * multi_edge_len_factor and edge_len > multi_edge_min_len:
                    deduped_graph.add_edge(s, e, geom=edge_geom)
                # otherwise, add to the list of longer geoms to be merged along with shortest
                else:
                    longer_geoms.append(edge_geom)
            # otherwise, if not merging on a midline basis
            # or, if no other edges to process (in cases where longer geom has been retained per above)
            # then use the shortest geom
            if not merge_edges_by_midline or len(longer_geoms) == 0:
                deduped_graph.add_edge(s, e, geom=shortest_geom)
            # otherwise weld the geoms, using the shortest as a yardstick
            else:
                # iterate the coordinates along the shorter geom
                new_coords = []
                for coord in shortest_geom.coords:
                    # from the current short_geom coordinate
                    short_point = geometry.Point(coord)
                    # find the nearest points on the longer geoms
                    multi_coords = [short_point]
                    for longer_geom in longer_geoms:
                        # get the nearest point on the longer geom
                        # returns a tuple of nearest geom for respective input geoms
                        longer_point = ops.nearest_points(short_point, longer_geom)[-1]
                        # aggregate
                        multi_coords.append(longer_point)
                    # create a midpoint between the geoms and add to the new coordinate array
                    mid_point = geometry.MultiPoint(multi_coords).centroid
                    new_coords.append(mid_point)
                # generate the new mid-line geom
                new_geom = geometry.LineString(new_coords)
                # add to the graph
                deduped_graph.add_edge(s, e, geom=new_geom)

    return deduped_graph


def _create_nodes_strtree(networkX_multigraph: nx.MultiGraph) -> strtree.STRtree:
    """
    Creates a nodes-based STRtree spatial index.
    """
    points = []
    for n, d in networkX_multigraph.nodes(data=True):
        # x coordinate
        if 'x' not in d:
            raise KeyError(f'Encountered node missing "x" coordinate attribute at node {n}.')
        x = d['x']
        # y coordinate
        if 'y' not in d:
            raise KeyError(f'Encountered node missing "y" coordinate attribute at node {n}.')
        y = d['y']
        p = geometry.Point(x, y)
        p.uid = n
        p.degree = nx.degree(networkX_multigraph, n)
        points.append(p)
    return strtree.STRtree(points)


def _create_edges_strtree(networkX_multigraph: nx.MultiGraph) -> strtree.STRtree:
    """
    Creates an edges-based STRtree spatial index.
    """
    lines = []
    for s, e, k, d in networkX_multigraph.edges(keys=True, data=True):
        if 'geom' not in d:
            raise KeyError(f'Encountered edge missing "geom" attribute.')
        linestring = d['geom']
        linestring.start_uid = s
        linestring.end_uid = e
        linestring.k = k
        lines.append(linestring)
    return strtree.STRtree(lines)


def nX_consolidate_nodes(networkX_multigraph: nx.MultiGraph,
                         buffer_dist: float = 5,
                         min_node_group: int = 2,
                         min_node_degree: int = 1,
                         min_cumulative_degree: int = None,
                         max_cumulative_degree: int = None,
                         neighbour_policy: str = None,
                         crawl: bool = True,
                         cent_min_degree: int = 3,
                         cent_min_len_factor: float = None,
                         merge_edges_by_midline: bool = True,
                         multi_edge_len_factor: float = 1.5,
                         multi_edge_min_len: float = 100) -> nx.MultiGraph:
    """
    Consolidates nodes if they are within a buffer distance of each other. Several parameters provide more control over
    the conditions used for deciding whether or not to merge nodes. The algorithm proceeds in two steps:
    - Nodes within the buffer distance of each other are merged. A new centroid will be determined and all existing
        edge endpoints will be updated accordingly. The new centroid for the merged nodes can be based on:
        - The centroid of the node group;
        - Else, all nodes of degree greater or equal to `cent_min_degree`;
        - Else, all nodes with aggregate adjacent edge lengths greater than a factor of `cent_min_len_factor` of the node
        with the greatest aggregate length for adjacent edges.
    - The merging of nodes creates parallel edges which may start and end at a shared node on either side. These edges
        are replaced by a single new edge, with the new geometry selected from either:
        - An imaginary centreline of the combined edges if `merge_edges_by_midline` is set to `True`;
        - Else, the shortest edge, with longer edges discarded;
        - Note that substantially longer parallel edges are retained, instead of discarded, if they exceed
          `multi_edge_len_factor` and are longer than `multi_edge_min_len`.

    Parameters
    ----------
    networkX_multigraph
        A `networkX` `MultiGraph` in a projected coordinate system, containing `x` and `y` node attributes, and `geom`
        edge attributes containing `LineString` geoms.
    buffer_dist
        The buffer distance to be used for consolidating nearby nodes. Defaults to 5.
    min_node_group
        The minimum number of nodes to consider a valid group for consolidation. Defaults to 2.
    min_node_degree
        The least number of edges a node should have in order to be considered for consolidation. Defaults to 1.
    min_cumulative_degree
        An optional minimum cumulative degree to consider a valid node group for consolidation. Defaults to None.
    max_cumulative_degree
        An optional maximum cumulative degree to consider a valid node group for consolidation. Defaults to None.
    neighbour_policy
        Whether all nodes within the buffer distance are merged, or only "direct" or "indirect" neighbours. Defaults to
        None.
    crawl
        Whether the algorithm will recursively explore neighbours of neighbours if those neighbours are within the
        buffer distance from the prior node. Defaults to True.
    cent_min_degree
        The minimum node degree for a node to be considered when calculating the new centroid for the merged node
        cluster. Defaults to 3.
    cent_min_len_factor
        The minimum aggregate adjacent edge lengths an existing node should have to be considered when calculating the
        centroid for the new node cluster. Expressed as a factor of the node with the greatest aggregate adjacent edge
        lengths. Defaults to None.
    merge_edges_by_midline
        Whether to merge parallel edges by an imaginary centreline. If set to False, then the shortest edge will be
        retained as the new geometry and the longer edges will be discarded. Defaults to True.
    multi_edge_len_factor
        In cases where one line is significantly longer than another (e.g. crescent streets) then the longer edge is
        retained as separate if exceeding the multi_edge_len_factor as a factor of the shortest length but with the
        exception that (longer) edges still shorter than multi_edge_min_len are removed regardless. Defaults to 1.5.
    multi_edge_min_len
        See `multi_edge_len_factor`. Defaults to 100.

    Returns
    -------
    nx.MultiGraph
        A `networkX` `MultiGraph` with consolidated nodes.

    Notes
    -----
    See the guide on [graph cleaning](/guide/cleaning) for more information.

    ![Example raw graph from OSM](../.vitepress/plots/images/graph_cleaning_1.png)
    _The pre-consolidation OSM street network for Soho, London. © OpenStreetMap contributors._

    ![Example cleaned graph](../.vitepress/plots/images/graph_cleaning_5.png)
    _The consolidated OSM street network for Soho, London. © OpenStreetMap contributors._
    """
    if not isinstance(networkX_multigraph, nx.MultiGraph):
        raise TypeError('This method requires an undirected networkX MultiGraph.')
    if min_node_group < 2:
        raise ValueError('The minimum node threshold should be set to at least two.')
    if neighbour_policy is not None and neighbour_policy not in ('direct', 'indirect'):
        raise ValueError('Neighbour policy should be one "direct", "indirect", else the default of "None"')
    if crawl and buffer_dist > 25:
        logger.warning('Be cautious with the buffer distance when using crawl.')
    logger.info('Consolidating nodes.')
    _multi_graph = networkX_multigraph.copy()
    # create a nodes STRtree
    nodes_tree = _create_nodes_strtree(_multi_graph)
    # keep track of removed nodes
    removed_nodes = set()

    def recursive_squash(nd_uid: Union[int, str],
                         x: float,
                         y: float,
                         node_group: list,
                         processed_nodes: list,
                         recursive: bool = False):
        # keep track of which nodes have been processed as part of recursion
        processed_nodes.append(nd_uid)
        # get all other nodes within buffer distance - the self-node and previously processed nodes are also returned
        js = nodes_tree.query(geometry.Point(x, y).buffer(buffer_dist))
        # review each node within the buffer
        for j in js:
            j_uid = j.uid
            if j_uid in removed_nodes or j_uid in processed_nodes or j.degree < min_node_degree:
                continue
            # check neighbour policy
            if neighbour_policy is not None:
                # use the original graph prior to in-place modifications
                neighbours = nx.neighbors(networkX_multigraph, nd_uid)
                if neighbour_policy == 'indirect' and j_uid in neighbours:
                    continue
                elif neighbour_policy == 'direct' and j_uid not in neighbours:
                    continue
            # otherwise add the node
            node_group.append(j_uid)
            # if recursive, follow the chain
            if recursive:
                j_nd = networkX_multigraph.nodes[j_uid]
                return recursive_squash(j_uid,
                                        j_nd['x'],
                                        j_nd['y'],
                                        node_group,
                                        processed_nodes,
                                        recursive=crawl)

        return node_group

    # iterate origin graph (else node structure changes in place)
    for n, n_d in tqdm(networkX_multigraph.nodes(data=True), disable=checks.quiet_mode):
        # skip if already consolidated from an adjacent node, or if the node's degree doesn't meet min_node_degree
        if n in removed_nodes or nx.degree(networkX_multigraph, n) < min_node_degree:
            continue
        node_group = recursive_squash(n,  # node uid
                                      n_d['x'],  # x point for buffer
                                      n_d['y'],  # y point for buffer
                                      [n],  # node group for consolidation (with starting node)
                                      [],  # processed nodes tracked through recursion
                                      crawl)  # whether to recursively probe neighbours per distance
        # check for min_node_threshold
        if len(node_group) < min_node_group:
            continue
        # check for cumulative degree thresholds if requested
        if min_cumulative_degree is not None or max_cumulative_degree is not None:
            cumulative_degree = sum([nx.degree(networkX_multigraph, n) for n in node_group])
            if min_cumulative_degree is not None and cumulative_degree < min_cumulative_degree:
                continue
            if max_cumulative_degree is not None and cumulative_degree > max_cumulative_degree:
                continue
        # update removed nodes
        removed_nodes.update(node_group)
        # consolidate if nodes have been identified within buffer and if these exceed min_node_threshold
        _multi_graph = _squash_adjacent(_multi_graph,
                                        node_group,
                                        cent_min_degree,
                                        cent_min_len_factor)
    # remove filler nodes
    deduped_graph = nX_remove_filler_nodes(_multi_graph)
    # remove any parallel edges that may have resulted from squashing nodes
    deduped_graph = _merge_parallel_edges(deduped_graph,
                                          merge_edges_by_midline,
                                          multi_edge_len_factor,
                                          multi_edge_min_len)

    return deduped_graph


def nX_split_opposing_geoms(networkX_multigraph: nx.MultiGraph,
                            buffer_dist: float = 10,
                            merge_edges_by_midline: bool = False,
                            multi_edge_len_factor: float = 1.5,
                            multi_edge_min_len: float = 100) -> nx.MultiGraph:
    """
    Projects nodes to pierce opposing edges within a buffer distance. The pierced nodes facilitate subsequent
    merging for scenarios such as divided boulevards.

    Parameters
    ----------
    networkX_multigraph
        A `networkX` `MultiGraph` in a projected coordinate system, containing `x` and `y` node attributes, and `geom`
        edge attributes containing `LineString` geoms.
    buffer_dist
        The buffer distance to be used for splitting nearby nodes. Defaults to 5.
    merge_edges_by_midline
        Whether to merge parallel edges by an imaginary centreline. If set to False, then the shortest edge will be
        retained as the new geometry and the longer edges will be discarded. Defaults to True.
    multi_edge_len_factor
        In cases where one line is significantly longer than another (e.g. crescent streets) then the longer edge is
        retained as separate if exceeding the multi_edge_len_factor as a factor of the shortest length but with the
        exception that (longer) edges still shorter than multi_edge_min_len are removed regardless. Defaults to 1.5.
    multi_edge_min_len
        See `multi_edge_len_factor`. Defaults to 100.

    Returns
    -------
    nx.MultiGraph
        A `networkX` `MultiGraph` with consolidated nodes.
    """

    def make_edge_key(s, e, k):
        return '-'.join(sorted([str(s), str(e)])) + f'-k{k}'

    # where edges are deleted, keep track of new children edges
    edge_children = {}

    # recursive function for retrieving nested layers of successively replaced edges
    def recurse_child_keys(s, e, k, geom, current_edges):
        """
        Checks if an edge has been replaced by children, if so, use children instead.
        Children may also have children, so recurse downwards.
        """
        edge_key = make_edge_key(s, e, k)
        # if an edge does not have children, add to current_edges and return
        if edge_key not in edge_children:
            current_edges.append((s, e, k, geom))
        # otherwise recursively drill-down until newest edges are found
        else:
            for child_s, child_e, child_k, child_geom in edge_children[edge_key]:
                recurse_child_keys(child_s, child_e, child_k, child_geom, current_edges)

    if not isinstance(networkX_multigraph, nx.MultiGraph):
        raise TypeError('This method requires an undirected networkX MultiGraph.')
    logger.info(f'Splitting opposing edges.')
    _multi_graph = networkX_multigraph.copy()
    # create an edges STRtree (nodes and edges)
    edges_tree = _create_edges_strtree(_multi_graph)
    # iterate origin graph (else node structure changes in place)
    for n, n_d in tqdm(networkX_multigraph.nodes(data=True), disable=checks.quiet_mode):
        # don't split opposing geoms from nodes of degree 1
        if nx.degree(_multi_graph, n) < 2:
            continue
        # get all other edges within the buffer distance
        # the spatial index using bounding boxes, so further filtering is required (see further down)
        # furthermore, successive iterations may remove old edges, so keep track of removed parent vs new child edges
        n_point = geometry.Point(n_d['x'], n_d['y'])
        # spatial query from point returns all buffers with buffer_dist
        edges = edges_tree.query(n_point.buffer(buffer_dist))
        # extract the start node, end node, geom
        edges = [(edge.start_uid, edge.end_uid, edge.k, edge) for edge in edges]
        # check against removed edges
        current_edges = []
        for s, e, k, edge_geom in edges:
            recurse_child_keys(s, e, k, edge_geom, current_edges)
        # get neighbouring nodes from new graph
        neighbours = list(_multi_graph.neighbors(n))
        # abort if only direct neighbours
        if len(current_edges) <= len(neighbours):
            continue
        # filter current_edges
        gapped_edges = []
        for s, e, k, edge_geom in current_edges:
            # skip direct neighbours
            if s == n or e == n:
                continue
            # check whether the geom is truly within the buffer distance
            if edge_geom.distance(n_point) > buffer_dist:
                continue
            gapped_edges.append((s, e, k, edge_geom))
        # abort if no gapped edges
        if not gapped_edges:
            continue
        # prepare the root node's point geom
        n_geom = geometry.Point(n_d['x'], n_d['y'])
        # iter gapped edges
        for s, e, k, edge_geom in gapped_edges:
            # see if start node is within buffer distance already
            s_nd_data = _multi_graph.nodes[s]
            s_nd_geom = geometry.Point(s_nd_data['x'], s_nd_data['y'])
            if s_nd_geom.distance(n_geom) <= buffer_dist:
                continue
            # likewise for end node
            e_nd_data = _multi_graph.nodes[e]
            e_nd_geom = geometry.Point(e_nd_data['x'], e_nd_data['y'])
            if e_nd_geom.distance(n_geom) <= buffer_dist:
                continue
            # otherwise, project a point and split the opposing geom
            # ops.nearest_points returns tuple of nearest from respective input geoms
            # want the nearest point on the line at index 1
            nearest_point = ops.nearest_points(n_geom, edge_geom)[-1]
            # if a valid nearest point has been found, go ahead and split the geom
            # use a snap because rounding precision errors will otherwise cause issues
            split_geoms = ops.split(ops.snap(edge_geom, nearest_point, 0.01), nearest_point)
            # in some cases the line will be pointing away, but is still near enough to be within max
            # in these cases a single geom will be returned
            if len(split_geoms) < 2:
                continue
            new_edge_geom_a, new_edge_geom_b = split_geoms
            # add the new node and edges to _multi_graph (don't modify networkX_multigraph because of iter in place)
            new_nd_name = _add_node(_multi_graph, [s, n, e], x=nearest_point.x, y=nearest_point.y)
            # if a node already exists at this location, add_node will return None
            if new_nd_name is None:
                continue
            _multi_graph.add_edge(s, new_nd_name)
            _multi_graph.add_edge(e, new_nd_name)
            if np.allclose(s_nd_geom.coords, new_edge_geom_a.coords[0][:2], atol=checks.tolerance, rtol=0) or \
                    np.allclose(s_nd_geom.coords, new_edge_geom_a.coords[-1][:2], atol=checks.tolerance, rtol=0):
                s_new_geom = new_edge_geom_a
                e_new_geom = new_edge_geom_b
            else:
                # double check matching geoms
                if not np.allclose(s_nd_geom.coords, new_edge_geom_b.coords[0][:2], atol=checks.tolerance, rtol=0) and \
                        not np.allclose(s_nd_geom.coords, new_edge_geom_b.coords[-1][:2], atol=checks.tolerance,
                                        rtol=0):
                    raise ValueError('Unable to match split geoms to existing nodes')
                s_new_geom = new_edge_geom_b
                e_new_geom = new_edge_geom_a
            # if splitting a looped component, then both new edges will have the same starting and ending nodes
            # in these cases, there will be multiple edges
            if s == e:
                assert _multi_graph.number_of_edges(s, new_nd_name) == 2
                s_k = 0
                e_k = 1
            else:
                assert _multi_graph.number_of_edges(s, new_nd_name) == 1
                assert _multi_graph.number_of_edges(e, new_nd_name) == 1
                s_k = e_k = 0
            # write the new edges
            _multi_graph[s][new_nd_name][s_k]['geom'] = s_new_geom
            _multi_graph[e][new_nd_name][e_k]['geom'] = e_new_geom
            # add the new edges to the edge_children dictionary
            edge_key = make_edge_key(s, e, k)
            edge_children[edge_key] = [(s, new_nd_name, s_k, s_new_geom),
                                       (e, new_nd_name, e_k, e_new_geom)]
            # drop the old edge from _multi_graph
            if _multi_graph.has_edge(s, e, k):
                _multi_graph.remove_edge(s, e, k)
    # squashing nodes can result in edge duplicates
    deduped_graph = _merge_parallel_edges(_multi_graph,
                                          merge_edges_by_midline,
                                          multi_edge_len_factor,
                                          multi_edge_min_len)

    return deduped_graph


def nX_decompose(networkX_multigraph: nx.MultiGraph,
                 decompose_max: float) -> nx.MultiGraph:
    """
    Decomposes a graph so that no edge is longer than a set maximum. Decomposition provides a more granular
    representation of potential variations along street lengths, while reducing network centrality side-effects that
    arise as a consequence of varied node densities.

    ::: warning Note
    Setting the `decompose` parameter too small in relation to the size of the graph may increase the computation time
    unnecessarily for subsequent analysis. For larger-scale urban analysis, it is generally not necessary to go smaller
    20m, and 50m may already be sufficient for the majority of cases.
    :::

    ::: tip Hint
    This function will automatically orient the `geom` attribute LineStrings in the correct direction before splitting
    into sub-geometries; i.e. there is no need to order the geometry's coordinates in a particular direction.
    :::

    Parameters
    ----------
    networkX_multigraph
        A `networkX` `MultiGraph` in a projected coordinate system, containing `x` and `y` node attributes, and `geom`
        edge attributes containing `LineString` geoms.

    decompose_max
        The maximum length threshold for decomposed edges.

    Returns
    -------
    nx.MultiGraph
        A decomposed `networkX` graph with no edge longer than the `decompose_max` parameter. If `live` node attributes
        were provided, then the `live` attribute for child-nodes will be set to `True` if either or both parent nodes
        were `live`. Otherwise, all nodes wil be set to `live=True`. The `length` and `impedance` edge attributes will
        be set to match the lengths of the new edges.

    Notes
    -----
    ```python
    from cityseer.tools import mock, graphs, plot

    G = mock.mock_graph()
    G_simple = graphs.nX_simple_geoms(G)
    G_decomposed = graphs.nX_decompose(G_simple, 100)
    plot.plot_nX(G_decomposed)
    ```

    ![Example graph](../.vitepress/plots/images/graph_simple.png)
    _Example graph prior to decomposition._

    ![Example decomposed graph](../.vitepress/plots/images/graph_decomposed.png)
    _Example graph after decomposition._

    """
    if not isinstance(networkX_multigraph, nx.MultiGraph):
        raise TypeError('This method requires an undirected networkX MultiGraph.')
    logger.info(f'Decomposing graph to maximum edge lengths of {decompose_max}.')
    g_multi_copy = networkX_multigraph.copy()
    # note -> write to a duplicated graph to avoid in-place errors
    for s, e, d in tqdm(networkX_multigraph.edges(data=True), disable=checks.quiet_mode):
        # test for x, y in start coordinates
        if 'x' not in networkX_multigraph.nodes[s] or 'y' not in networkX_multigraph.nodes[s]:
            raise KeyError(f'Encountered node missing "x" or "y" coordinate attributes at node {s}.')
        # test for x, y in end coordinates
        if 'x' not in networkX_multigraph.nodes[e] or 'y' not in networkX_multigraph.nodes[e]:
            raise KeyError(f'Encountered node missing "x" or "y" coordinate attributes at node {e}.')
        s_x = networkX_multigraph.nodes[s]['x']
        s_y = networkX_multigraph.nodes[s]['y']
        e_x = networkX_multigraph.nodes[e]['x']
        e_y = networkX_multigraph.nodes[e]['y']
        # test for geom
        if 'geom' not in d:
            raise KeyError(
                f'No edge geom found for edge {s}-{e}: '
                f'Please add an edge "geom" attribute consisting of a shapely LineString.')
        # get edge geometry
        line_geom = d['geom']
        if line_geom.type != 'LineString':
            raise TypeError(f'Expecting LineString geometry but found {line_geom.type} geometry for edge {s}-{e}.')
        # check geom coordinates directionality - flip if facing backwards direction
        line_geom_coords = _align_linestring_coords(line_geom.coords, (s_x, s_y))
        # double check that coordinates now face the forwards direction
        if not np.allclose((s_x, s_y), line_geom_coords[0][:2], atol=checks.tolerance, rtol=0) or \
                not np.allclose((e_x, e_y), line_geom_coords[-1][:2], atol=checks.tolerance, rtol=0):
            raise ValueError(f'Edge geometry endpoint coordinate mismatch for edge {s}-{e}')
        line_geom = geometry.LineString(line_geom_coords)
        # see how many segments are necessary so as not to exceed decomposition max distance
        # note that a length less than the decompose threshold will result in a single 'sub'-string
        n = np.ceil(line_geom.length / decompose_max)
        step_size = line_geom.length / n
        # since decomposing, remove the prior edge... but only after properties have been read
        g_multi_copy.remove_edge(s, e)
        # then add the new sub-edge/s
        step = 0
        prior_node_id = s
        sub_node_counter = 0
        # everything inside this loop is a new node - i.e. this loop is effectively skipped if n = 1
        for i in range(int(n) - 1):
            # create the split LineString geom for measuring the new length
            line_segment = ops.substring(line_geom, step, step + step_size)
            # get the x, y of the new end node
            x, y = line_segment.coords[-1]
            # add the new node and edge
            new_nd_name = _add_node(g_multi_copy, [s, sub_node_counter, e], x=x, y=y)
            if new_nd_name is None:
                raise ValueError(f'Attempted to add a duplicate node. '
                                 f'Check for existence of duplicate edges in the vicinity of {s}-{e}.')
            sub_node_counter += 1
            # add and set live property if present in parent graph
            if 'live' in networkX_multigraph.nodes[s] and 'live' in networkX_multigraph.nodes[e]:
                live = True
                # if BOTH parents are not live, then set child to not live
                if not networkX_multigraph.nodes[s]['live'] and not networkX_multigraph.nodes[e]['live']:
                    live = False
                g_multi_copy.nodes[new_nd_name]['live'] = live
            # add the edge
            g_multi_copy.add_edge(prior_node_id, new_nd_name, geom=line_segment)
            # increment the step and node id
            prior_node_id = new_nd_name
            step += step_size
        # set the last edge manually to avoid rounding errors at end of LineString
        # the nodes already exist, so just add edge
        line_segment = ops.substring(line_geom, step, line_geom.length)
        g_multi_copy.add_edge(prior_node_id, e, geom=line_segment)

    return g_multi_copy


def nX_to_dual(networkX_multigraph: nx.MultiGraph) -> nx.MultiGraph:
    """
    Converts a primal graph representation, where intersections are represented as nodes and streets as edges, to the
    dual representation. So doing, edges are converted to nodes and intersections become edges. Primal edge `geom`
    attributes will be welded to adjacent edges and split into the new dual edge `geom` attributes.

    :::tip Note
    Note that a `MultiGraph` is useful for primal but not for dual, so the output `MultiGraph` will have single edges.
    e.g. a crescent street that spans the same intersections as parallel straight street requires multiple edges in
    primal. The same type of situation does not arise in the dual because the nodes map to distinct streets regardless.
    :::

    Parameters
    ----------
    networkX_multigraph
        A `networkX` `MultiGraph` in a projected coordinate system, containing `x` and `y` node attributes, and `geom`
        edge attributes containing `LineString` geoms.

    Returns
    -------
    nx.MultiGraph
        A dual representation `networkX` graph. The new dual nodes will have `x` and `y` node attributes corresponding
        to the mid-points of the original primal edges. If `live` node attributes were provided, then the `live`
        attribute for the new dual nodes will be set to `True` if either or both of the adjacent primal nodes were set
        to `live=True`. Otherwise, all dual nodes wil be set to `live=True`. The primal `geom` edge attributes will be
        split and welded to form the new dual `geom` edge attributes. A `parent_primal_node` edge attribute will be
        added, corresponding to the node identifier of the primal graph.

    Notes
    -----
    ```python
    from cityseer.tools import graphs, mock, plot

    G = mock.mock_graph()
    G_simple = graphs.nX_simple_geoms(G)
    G_dual = graphs.nX_to_dual(G_simple)
    plot.plot_nX_primal_or_dual(G_simple,
                                G_dual,
                                plot_geoms=False)
    ```

    ![Example dual graph](../.vitepress/plots/images/graph_dual.png)
    _Dual graph (blue) overlaid on the source primal graph (red)._

    """

    if not isinstance(networkX_multigraph, nx.MultiGraph):
        raise TypeError('This method requires an undirected networkX MultiGraph.')
    logger.info('Converting graph to dual.')
    g_dual = nx.MultiGraph()

    def get_half_geoms(g, a_node, b_node, edge_k):
        '''
        For splitting and orienting half geoms
        '''
        # get edge data
        edge_data = g[a_node][b_node][edge_k]
        # test for x coordinates
        if 'x' not in g.nodes[a_node] or 'y' not in g.nodes[a_node]:
            raise KeyError(f'Encountered node missing "x" or "y" coordinate attributes at node {a_node}.')
        # test for y coordinates
        if 'x' not in g.nodes[b_node] or 'y' not in g.nodes[b_node]:
            raise KeyError(f'Encountered node missing "x" or "y" coordinate attributes at node {b_node}.')
        a_xy = (g.nodes[a_node]['x'], g.nodes[a_node]['y'])
        b_xy = (g.nodes[b_node]['x'], g.nodes[b_node]['y'])
        # test for geom
        if 'geom' not in edge_data:
            raise KeyError(
                f'No edge geom found for edge {a_node}-{b_node}: '
                f'Please add an edge "geom" attribute consisting of a shapely LineString.')
        # get edge geometry
        line_geom = edge_data['geom']
        if line_geom.type != 'LineString':
            raise TypeError(
                f'Expecting LineString geometry but found {line_geom.type} geometry for edge {a_node}-{b_node}.')
        # align geom coordinates to start from A side
        line_geom_coords = _align_linestring_coords(line_geom.coords, a_xy)
        line_geom = geometry.LineString(line_geom_coords)
        # generate the two half geoms
        a_half_geom = ops.substring(line_geom, 0, line_geom.length / 2)
        b_half_geom = ops.substring(line_geom, line_geom.length / 2, line_geom.length)
        # check that nothing odd happened with new midpoint
        assert np.allclose(a_half_geom.coords[-1][:2], b_half_geom.coords[0][:2], atol=checks.tolerance, rtol=0)
        # snap to prevent creeping tolerance issues
        # A side geom starts at node A and ends at new midpoint
        a_half_geom_coords = _snap_linestring_startpoint(a_half_geom.coords, a_xy)
        # snap new midpoint to geom A's endpoint (i.e. no need to snap endpoint of geom A)
        mid_xy = a_half_geom_coords[-1][:2]
        # B side geom starts at mid and ends at B node
        b_half_geom_coords = _snap_linestring_startpoint(b_half_geom.coords, mid_xy)
        b_half_geom_coords = _snap_linestring_endpoint(b_half_geom_coords, b_xy)
        # double check coords
        assert a_half_geom_coords[0][:2] == a_xy
        assert a_half_geom_coords[-1][:2] == mid_xy
        assert b_half_geom_coords[0][:2] == mid_xy
        assert b_half_geom_coords[-1][:2] == b_xy

        return geometry.LineString(a_half_geom_coords), geometry.LineString(b_half_geom_coords)

    def set_live(s, e, dual_n):
        # add and set live property to dual node if present in adjacent primal graph nodes
        if 'live' in networkX_multigraph.nodes[s] and 'live' in networkX_multigraph.nodes[e]:
            live = True
            # if BOTH parents are not live, then set child to not live
            if not networkX_multigraph.nodes[s]['live'] and not networkX_multigraph.nodes[e]['live']:
                live = False
            g_dual.nodes[dual_n]['live'] = live

    # iterate the primal graph's edges
    for s, e, k, d in tqdm(networkX_multigraph.edges(data=True, keys=True), disable=checks.quiet_mode):
        # get the first and second half geoms
        s_half_geom, e_half_geom = get_half_geoms(networkX_multigraph, s, e, k)
        # create a new dual node corresponding to the current primal edge
        # nodes are added manually to retain link to origin node names and to check for duplicates
        s_e = sorted([str(s), str(e)])
        hub_node_dual = f'{s_e[0]}_{s_e[1]}'
        # the node may already have been added from a neighbouring node that has already been processed
        if hub_node_dual not in g_dual:
            x, y = s_half_geom.coords[-1][:2]
            g_dual.add_node(hub_node_dual, x=x, y=y)
            # add and set live property if present in parent graph
            set_live(s, e, hub_node_dual)
        # process either side
        for n_side, half_geom in zip([s, e], [s_half_geom, e_half_geom]):
            # add the spoke edges on the dual
            for nb in nx.neighbors(networkX_multigraph, n_side):
                # don't follow neighbour back to current edge combo
                if nb in [s, e]:
                    continue
                # add the neighbouring primal edge as dual node
                s_nb = sorted([str(n_side), str(nb)])
                spoke_node_dual = f'{s_nb[0]}_{s_nb[1]}'
                # skip if the edge has already been processed from another direction
                if g_dual.has_edge(hub_node_dual, spoke_node_dual):
                    continue
                # get the near and far half geoms
                spoke_half_geom, _discard_geom = get_half_geoms(networkX_multigraph, n_side, nb, k)
                # nodes will be added if not already present (i.e. from first direction processed)
                if spoke_node_dual not in g_dual:
                    x, y = spoke_half_geom.coords[-1][:2]
                    g_dual.add_node(spoke_node_dual, x=x, y=y)
                    # add and set live property if present in parent graph
                    set_live(s, e, spoke_node_dual)
                # weld the lines
                merged_line = ops.linemerge([half_geom, spoke_half_geom])
                if merged_line.type != 'LineString':
                    raise TypeError(
                        f'Found {merged_line.type} geometry instead of "LineString" for new geom {merged_line.wkt}. '
                        f'Check that the LineStrings for {s}-{e} and {n_side}-{nb} actually touch.')
                # add the dual edge
                g_dual.add_edge(hub_node_dual, spoke_node_dual, parent_primal_node=n_side, geom=merged_line)

    return g_dual


def graph_maps_from_nX(networkX_multigraph: nx.MultiGraph) -> Tuple[tuple, np.ndarray, np.ndarray, Dict]:
    """
    Transposes a `networkX` `MultiGraph` into `numpy` arrays for use by `NetworkLayer` classes. Calculates length and
    angle attributes, as well as in and out bearings and stores these in the returned data maps.

    ::: warning Note
    It is generally not necessary to use this function directly. This function will be called internally when invoking
    [NetworkLayerFromNX](/metrics/networks#class-networklayerfromnx)
    :::

    Parameters
    ----------
    networkX_multigraph
        A `networkX` `MultiGraph` in a projected coordinate system, containing `x` and `y` node attributes, and `geom`
        edge attributes containing `LineString` geoms.

    Returns
    -------
    node_uids
        A tuple of node `uids` corresponding to the node identifiers in the source `networkX` graph.
    node_data
        A 2d `numpy` array representing the graph's nodes. The indices of the second dimension correspond as follows:
        | idx | property |
        |-----|:---------|
        | 0 | `x` coordinate |
        | 1 | `y` coordinate |
        | 2 | `bool` describing whether the node is `live`. Metrics are only computed for `live` nodes. |
    edge_data
        A 2d `numpy` array representing the graph's edges. Each edge will be described separately for each direction of
        travel. The indices of the second dimension correspond as follows:
        | idx | property |
        |-----|:---------|
        | 0 | start node `idx` |
        | 1 | end node `idx` |
        | 2 | the segment length in metres |
        | 3 | the sum of segment's angular change |
        | 4 | an 'impedance factor' which can be applied to magnify or reduce the effect of the edge's impedance on
        shortest-path calculations. e.g. for gradients or other such considerations. Use with caution. |
        | 5 | the edge's entry angular bearing |
        | 6 | the edge's exit angular bearing |
        All edge attributes will be generated automatically, however, the impedance factor parameter can be over-ridden by supplying a `imp_factor` attribute on the input graph's edges.
    node_edge_map
        A `numba` `Dict` with `node_data` indices as keys and `numba` `List` types as values containing the out-edge
        indices for each node.
    """

    if not isinstance(networkX_multigraph, nx.MultiGraph):
        raise TypeError('This method requires an undirected networkX MultiGraph.')
    logger.info('Preparing node and edge arrays from networkX graph.')
    g_multi_copy = networkX_multigraph.copy()
    # accumulate degrees
    total_out_degrees = 0
    for n in tqdm(g_multi_copy.nodes(), disable=checks.quiet_mode):
        # writing node identifier to 'labels' in case conversion to integers method interferes with order
        g_multi_copy.nodes[n]['label'] = n
        for nb in nx.neighbors(g_multi_copy, n):
            total_out_degrees += g_multi_copy.number_of_edges(n, nb)
    # convert the nodes to sequential - this permits implicit indices with benefits to speed and structure
    g_multi_copy = nx.convert_node_labels_to_integers(g_multi_copy, 0)
    # prepare the node and edge maps
    node_uids = []
    # float - for consistency - requires higher accuracy for x, y work
    node_data = np.full((g_multi_copy.number_of_nodes(), 3), np.nan, dtype=np.float64)
    # float - allows for nan and inf - float32 should be ample...
    edge_data = np.full((total_out_degrees, 7), np.nan, dtype=np.float32)
    # nodes have a one-to-many mapping to edges
    node_edge_map = Dict.empty(
        key_type=types.int64,
        value_type=types.int64[:]
    )
    edge_idx = 0
    # populate the nodes
    for n, d in tqdm(g_multi_copy.nodes(data=True), disable=checks.quiet_mode):
        # label
        # don't cast to string because otherwise correspondence between original and round-trip graph indices is lost
        node_uids.append(d['label'])
        # cast to int for indexing
        node_idx = int(n)
        # NODE MAP INDEX POSITION 0 = x coordinate
        if 'x' not in d:
            raise KeyError(f'Encountered node missing "x" coordinate attribute at node {n}.')
        node_data[node_idx][0] = d['x']
        # NODE MAP INDEX POSITION 1 = y coordinate
        if 'y' not in d:
            raise KeyError(f'Encountered node missing "y" coordinate attribute at node {n}.')
        node_data[node_idx][1] = d['y']
        # NODE MAP INDEX POSITION 2 = live or not
        if 'live' in d:
            node_data[node_idx][2] = d['live']
        else:
            node_data[node_idx][2] = True
        # build edges
        out_edges = []
        for nb in g_multi_copy.neighbors(n):
            for nx_edge_idx, nx_edge_data in g_multi_copy[n][nb].items():
                # add the new edge index to the node's out edges
                out_edges.append(edge_idx)
                # EDGE MAP INDEX POSITION 0 = start node
                edge_data[edge_idx][0] = node_idx
                # EDGE MAP INDEX POSITION 1 = end node
                edge_data[edge_idx][1] = nb
                # EDGE MAP INDEX POSITION 2 = length
                if not 'geom' in nx_edge_data:
                    raise KeyError(
                        f'No edge geom found for edge {node_idx}-{nb}: '
                        f'Please add an edge "geom" attribute consisting of a shapely LineString.'
                        f'Simple (straight) geometries can be inferred automatically through use of the nX_simple_geoms() method.')
                line_geom = nx_edge_data['geom']
                if line_geom.type != 'LineString':
                    raise TypeError(
                        f'Expecting LineString geometry but found {line_geom.type} geometry for edge {node_idx}-{nb}.')
                # cannot have zero or negative length - division by zero
                l = line_geom.length
                if not np.isfinite(l) or l <= 0:
                    raise ValueError(f'Length attribute {l} for edge {node_idx}-{nb} must be a finite positive value.')
                edge_data[edge_idx][2] = l
                # EDGE MAP INDEX POSITION 3 = angle_sum
                # check geom coordinates directionality (for bearings at index 5 / 6)
                # flip if facing backwards direction
                s_x, s_y = node_data[node_idx][:2]
                line_geom_coords = _align_linestring_coords(line_geom.coords, (s_x, s_y))
                # iterate the coordinates and calculate the angular change
                angle_sum = 0
                for c in range(len(line_geom_coords) - 2):
                    x_1, y_1 = line_geom_coords[c][:2]
                    x_2, y_2 = line_geom_coords[c + 1][:2]
                    x_3, y_3 = line_geom_coords[c + 2][:2]
                    # arctan2 is y / x order
                    a_1 = np.rad2deg(np.arctan2(y_2 - y_1, x_2 - x_1))
                    a_2 = np.rad2deg(np.arctan2(y_3 - y_2, x_3 - x_2))
                    angle_sum += np.abs((a_2 - a_1 + 180) % 360 - 180)
                    # alternative
                    # A = np.array(merged_line.coords[c + 1]) - np.array(merged_line.coords[c])
                    # B = np.array(merged_line.coords[c + 2]) - np.array(merged_line.coords[c + 1])
                    # angle = np.abs(np.degrees(np.math.atan2(np.linalg.det([A, B]), np.dot(A, B))))
                if not np.isfinite(angle_sum) or angle_sum < 0:
                    raise ValueError(
                        f'Angle-sum attribute {angle_sum} for edge {node_idx}-{nb} must be a finite positive value.')
                edge_data[edge_idx][3] = angle_sum
                # EDGE MAP INDEX POSITION 4 = imp_factor
                # if imp_factor is set explicitly, then use
                if 'imp_factor' in nx_edge_data:
                    # cannot have imp_factor less than zero (but == 0 is OK)
                    imp_factor = nx_edge_data['imp_factor']
                    if not (np.isfinite(imp_factor) or np.isinf(imp_factor)) or imp_factor < 0:
                        raise ValueError(
                            f'Impedance factor: {imp_factor} for edge {node_idx}-{nb} must be a finite positive value or positive infinity.')
                    edge_data[edge_idx][4] = imp_factor
                else:
                    # fallback imp_factor of 1
                    edge_data[edge_idx][4] = 1
                # EDGE MAP INDEX POSITION 5 - in bearing
                x_1, y_1 = line_geom_coords[0][:2]
                x_2, y_2 = line_geom_coords[1][:2]
                edge_data[edge_idx][5] = np.rad2deg(np.arctan2(y_2 - y_1, x_2 - x_1))
                # EDGE MAP INDEX POSITION 6 - out bearing
                x_1, y_1 = line_geom_coords[-2][:2]
                x_2, y_2 = line_geom_coords[-1][:2]
                edge_data[edge_idx][6] = np.rad2deg(np.arctan2(y_2 - y_1, x_2 - x_1))
                # increment the edge_idx
                edge_idx += 1
        # add the node to the node_edge_map
        node_edge_map[node_idx] = np.array(out_edges, dtype='int64')

    return tuple(node_uids), node_data, edge_data, node_edge_map


def nX_from_graph_maps(node_uids: Union[tuple, list],
                       node_data: np.ndarray,
                       edge_data: np.ndarray,
                       node_edge_map: Dict,
                       networkX_multigraph: nx.MultiGraph = None,
                       metrics_dict: dict = None) -> nx.MultiGraph:
    """
    Writes cityseer data graph maps back to a `MultiGraph`. Can write back to an existing `MultiGraph` if an existing
    graph is provided as an argument to the `networkX_multigraph` parameter.

    ::: warning Note
    It is generally not necessary to use this function directly. This function will be called internally when invoking
    [NetworkLayer.to_networkX](/metrics/networks#to-networkX)
    :::

    Parameters
    ----------
    node_uids
        A tuple of node ids corresponding to the node identifiers for the target `networkX` graph.
    node_data
        A 2d `numpy` array representing the graph's nodes. The indices of the second dimension should correspond as
        follows:
        | idx | property |
        | --- | :------- |
        | 0   | `x` coordinate |
        | 1   | `y` coordinate |
        | 2   | `bool` describing whether the node is `live` |
    edge_data
        A 2d `numpy` array representing the graph's directional edges. The indices of the second dimension should 
        correspond as follows:
        | idx | property |
        | --- | :------- |
        | 0   | start node `idx` |
        | 1   | end node `idx` |
        | 2   | the segment length in metres |
        | 3   | the sum of segment's angular change |
        | 4   | 'impedance factor' applied to magnify or reduce the edge impedance. |
        | 5   | the edge's entry angular bearing |
        | 6   | the edge's exit angular bearing |
    node_edge_map
        A `numba` `Dict` with `node_data` indices as keys and `numba` `List` types as values containing the out-edge
        indices for each node.
    networkX_multigraph
        An optional `networkX` graph to use as a backbone for unpacking the data. The number of nodes and edges should
        correspond to the `cityseer` data maps and the node identifiers should correspond to the `node_uids`. If not
        provided, then a new `networkX` graph will be returned. This function is intended to be used for situations
        where `cityseer` data is being transposed back to a source `networkX` graph. Defaults to None.
    metrics_dict
        An optional dictionary with keys corresponding to the identifiers in `node_uids`. The dictionary's `values` will
        be unpacked to the corresponding nodes in the `networkX` graph. Defaults to None.

    Returns
    -------
    nx.MultiGraph
        A `networkX` graph. If a backbone graph was provided, a copy of the same graph will be returned with the data
        overridden as described below. If no graph was provided, then a new graph will be generated.
        `x`, `y`, `live`, `ghosted` node attributes will be copied from `node_data` to the graph nodes. `length`,
        `angle_sum`, `imp_factor`, `start_bearing`, and `end_bearing` attributes will be copied from the `edge_data`
        to the graph edges. If a `metrics_dict` is provided, all data will be copied to the graph nodes based on
        matching node identifiers.
    """
    logger.info('Populating node and edge map data to a networkX graph.')
    if networkX_multigraph is not None:
        logger.info('Reusing existing graph as backbone.')
        if networkX_multigraph.number_of_nodes() != len(node_data):
            raise ValueError('The number of nodes in the graph does not match the number of nodes in the node map.')
        g_multi_copy = networkX_multigraph.copy()
        for uid in node_uids:
            if uid not in g_multi_copy:
                raise KeyError(
                    f'Node uid {uid} not found in graph. '
                    f'If passing a graph as backbone, the uids must match those supplied with the node and edge maps.')
    else:
        logger.info('No existing graph found, creating new.')
        g_multi_copy = nx.MultiGraph()
        g_multi_copy.add_nodes_from(node_uids)
    # after above so that errors caught first
    checks.check_network_maps(node_data, edge_data, node_edge_map)
    logger.info('Unpacking node data.')
    for uid, node in tqdm(zip(node_uids, node_data), disable=checks.quiet_mode):
        x, y, live = node
        g_multi_copy.nodes[uid]['x'] = x
        g_multi_copy.nodes[uid]['y'] = y
        g_multi_copy.nodes[uid]['live'] = bool(live)
    logger.info('Unpacking edge data.')
    for edge in tqdm(edge_data, disable=checks.quiet_mode):
        start, end, length, angle_sum, imp_factor, start_bearing, end_bearing = edge
        start_uid = node_uids[int(start)]
        end_uid = node_uids[int(end)]
        # note that the original geom is lost with round trip unless retained in a supplied backbone graph.
        # the edge map is directional, so each original edge will be processed twice, once from either direction.
        # edges are only added if A) not using a backbone graph and B) the edge hasn't already been added
        if networkX_multigraph is None:
            # if the edge doesn't already exist, then simply add
            if not g_multi_copy.has_edge(start_uid, end_uid):
                add_edge = True
            # else, only add if not matching an already added edge
            # i.e. don't add the same edge when processed from opposite direction
            else:
                add_edge = True  # tentatively set to True
                # iter the edges
                for edge_item_idx, edge_item_data in g_multi_copy[start_uid][end_uid].items():
                    # set add_edge to false if a matching edge length is found
                    if edge_item_data['length'] == length:
                        add_edge = False
            # add the edge if not existent
            if add_edge:
                g_multi_copy.add_edge(start_uid,
                                      end_uid,
                                      length=length,
                                      angle_sum=angle_sum,
                                      imp_factor=imp_factor)
        # if a new edge is not being added then add the attributes to the appropriate backbone edge if not already done
        # this is only relevant if processing a backbone graph
        else:
            # raise if the edge doesn't exist
            if not g_multi_copy.has_edge(start_uid, end_uid):
                raise KeyError(f'The backbone graph is missing an edge spanning from {start_uid} to {end_uid}'
                               f'The original graph (with all original edges) has to be reused.')
            # due working with a MultiGraph it is necessary to check that the correct edge index is matched
            for edge_item_idx, edge_item_data in g_multi_copy[start_uid][end_uid].items():
                if np.isclose(edge_item_data['geom'].length, length, atol=0.1, rtol=0.0):
                    # check whether the attributes have already been added from the other direction?
                    if 'length' in edge_item_data and edge_item_data['length'] == length:
                        continue
                    # otherwise add the edge attributes and move on
                    existing_edge = g_multi_copy[start_uid][end_uid][edge_item_idx]
                    existing_edge['length'] = length
                    existing_edge['angle_sum'] = angle_sum
                    existing_edge['imp_factor'] = imp_factor
    # unpack any metrics written to the nodes
    if metrics_dict is not None:
        logger.info('Unpacking metrics to nodes.')
        for uid, metrics in tqdm(metrics_dict.items(), disable=checks.quiet_mode):
            if uid not in g_multi_copy:
                raise KeyError(
                    f'Node uid {uid} not found in graph. '
                    f'Data dictionary uids must match those supplied with the node and edge maps.')
            g_multi_copy.nodes[uid]['metrics'] = metrics

    return g_multi_copy



import os
import numpy as np
from numba import njit
from numba.typed import Dict

# for calculating default betas vs. distances
def_min_thresh_wt = 0.01831563888873418
# for checking linestring geometries
tolerance = 0.001

quiet_mode = False
if 'GCP_PROJECT' in os.environ:
    quiet_mode = True

if 'CITYSEER_QUIET_MODE' in os.environ:
    if os.environ['CITYSEER_QUIET_MODE'].lower() in ['true', '1']:
        quiet_mode = True

debug_mode = False
if 'CITYSEER_DEBUG_MODE' in os.environ:
    if os.environ['CITYSEER_DEBUG_MODE'].lower() in ['true', '1']:
        debug_mode = True


@njit(cache=True)
def _print_msg(hash_count, void_count, percentage):
    msg = '|'
    for n in range(int(hash_count)):
        msg += '#'
    for n in range(int(void_count)):
        msg += ' '
    msg += '|'
    print(msg, percentage, '%')


@njit(cache=False)
def progress_bar(current: int, total: int, steps: int = 20):
    '''
    Printing carries a performance penalty
    Cache has to be set to false per Numba issue:
    https://github.com/numba/numba/issues/3555
    TODO: set cache to True once resolved - likely 2020
    '''
    if steps == 0:
        return
    if current + 1 == total:
        _print_msg(steps, 0, 100)
        return
    if total <= steps:
        step_size = 1
    else:
        step_size = int(total / steps)
    if current % step_size == 0:
        percentage = np.round(current / total * 100, 2)
        hash_count = int(percentage / 100 * steps)
        void_count = steps - hash_count
        _print_msg(hash_count, void_count, percentage)


@njit(cache=True)
def check_numerical_data(data_arr: np.ndarray):
    if not data_arr.ndim == 2:
        raise ValueError('The numeric data array must have a dimensionality 2, '
                         'consisting of the number of respective data arrays x the length of data points.')
    for num in np.nditer(data_arr):
        if np.isinf(num):
            raise ValueError('The numeric data values must consist of either floats or NaNs.')


@njit(cache=True)
def check_categorical_data(data_arr: np.ndarray):
    for cl in data_arr:
        if not np.isfinite(np.float(cl)) or not cl >= 0:
            raise ValueError('Data map contains points with missing data classes.')
        if int(cl) != cl:
            raise ValueError('Data map contains non-integer class-codes.')


@njit(cache=True)
def check_data_map(data_map: np.ndarray, check_assigned=True):
    '''
    DATA MAP:
    0 - x
    1 - y
    2 - assigned network index - nearest
    3 - assigned network index - next-nearest
    '''
    # catch zero length data maps
    if len(data_map) == 0:
        raise ValueError('Zero length data map')

    # other checks - e.g. checking for single dimensional arrays, are tricky with numba
    if not data_map.ndim == 2 or not data_map.shape[1] == 4:
        raise ValueError(
            'The data map must have a dimensionality of Nx4, with the first two indices consisting of x, y coordinates. '
            'Indices 2 and 3, if populated, correspond to the nearest and next-nearest network nodes.')

    if check_assigned:
        # check that data map has been assigned
        if np.all(np.isnan(data_map[:, 2])):
            raise ValueError('Data map has not been assigned to a network.')


@njit(cache=True)
def check_network_maps(node_data: np.ndarray,
                       edge_data: np.ndarray,
                       node_edge_map: Dict):
    '''
    NODE MAP:
    0 - x
    1 - y
    2 - live
    EDGE MAP:
    0 - start node
    1 - end node
    2 - length in metres
    3 - sum of angular travel along length
    4 - impedance factor
    5 - entry bearing
    6 - exit bearing
    '''
    # catch zero length node or edge maps
    if len(node_data) == 0:
        raise ValueError('Zero length node map')
    if len(edge_data) == 0:
        raise ValueError('Zero length edge map')
    if not node_data.ndim == 2 or not node_data.shape[1] == 3:
        raise ValueError('The node map must have a dimensionality of Nx4.')
    if not edge_data.ndim == 2 or not edge_data.shape[1] == 7:
        raise ValueError('The edge map must have a dimensionality of Nx7')
    # check sequential and reciprocal node to edge map indices
    edge_counts = np.full(len(edge_data), 0)
    for n_idx in range(len(node_data)):
        # zip through all edges for current node
        for edge_idx in node_edge_map[n_idx]:
            # get the edge
            edge = edge_data[edge_idx]
            # check that the start node matches the current node index
            start_nd_idx, end_nd_idx = edge[:2]
            assert start_nd_idx == n_idx
            # check that each edge has a matching pair in the opposite direction
            paired = False
            for return_edge_idx in node_edge_map[int(end_nd_idx)]:
                if edge_data[return_edge_idx][1] == n_idx:
                    paired = True
                    break
            if not paired:
                raise ValueError('Missing matching edge pair in opposite direction.')
            # add to the counter
            edge_counts[edge_idx] += 1
    if not np.all(edge_counts == 1):
        raise ValueError('Mismatched node and edge maps encountered.')
    if not np.all(np.isfinite(edge_data[:, 0])) or not np.all(edge_data[:, 0] >= 0):
        raise ValueError('Missing or invalid start node index encountered.')
    if not np.all(np.isfinite(edge_data[:, 1])) or not np.all(edge_data[:, 1] >= 0):
        raise ValueError('Missing or invalid end node index encountered.')
    if not np.all(np.isfinite(edge_data[:, 2])) or not np.all(edge_data[:, 2] >= 0):
        raise ValueError('Invalid edge length encountered. Should be finite number greater than or equal to zero.')
    if not np.all(np.isfinite(edge_data[:, 3])) or not np.all(edge_data[:, 3] >= 0):
        raise ValueError(
            'Invalid edge angle sum encountered. Should be finite number greater than or equal to zero.')
    if not np.all(np.isfinite(edge_data[:, 4])) or not np.all(edge_data[:, 4] >= 0):
        raise ValueError(
            'Invalid impedance factor encountered. Should be finite number greater than or equal to zero.')


@njit(cache=True)
def check_distances_and_betas(distances: np.ndarray, betas: np.ndarray):
    if len(distances) == 0:
        raise ValueError('No distances provided.')

    if len(betas) == 0:
        raise ValueError('No betas provided.')

    if not len(distances) == len(betas):
        raise ValueError('The number of distances and betas should be equal.')

    for i in range(len(distances)):
        for j in range(len(distances)):
            if i > j:
                if distances[i] == distances[j]:
                    raise ValueError('Duplicate distances provided. Please provide only one of each.')

    for d in distances:
        if d <= 0:
            raise ValueError('Please provide a positive distance value.')

    for b in betas:
        if b > 0:
            raise ValueError('Please provide the beta value with the leading negative.')

    threshold_min = np.exp(distances[0] * betas[0])
    for d, b in zip(distances, betas):
        if not np.exp(d * b) == threshold_min:
            # handles edge case for infinity
            if not (d, b) == (np.inf, -0.0):
                raise ValueError(
                    'Inconsistent threshold minimums, indicating that the relationship between the betas and distances '
                    'is not consistent for all distance / beta pairs.')



from typing import Tuple
import numpy as np
from numba import njit, types
from numba.typed import List, Dict

from cityseer.algos import checks


# don't use 'nnan' fastmath flag
@njit(cache=True, fastmath={'ninf', 'nsz', 'arcp', 'contract', 'afn', 'reassoc'})
def shortest_path_tree(
        edge_data: np.ndarray,
        node_edge_map: Dict,
        src_idx: int,
        max_dist: float = np.inf,
        angular: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    '''
    All shortest paths to max network distance from source node
    Returns impedances and predecessors for shortest paths from a source node to all other nodes within max distance
    Angular flag triggers check for sidestepping / cheating with angular impedances (sharp turns)

    NODE MAP:
    0 - x
    1 - y
    2 - live

    EDGE MAP:
    0 - start node
    1 - end node
    2 - length in metres
    3 - sum of angular travel along length
    4 - impedance factor
    5 - in bearing
    6 - out bearing

    RETURNS A SHORTEST PATH TREE MAP:
    0 - processed nodes
    1 - predecessors
    2 - distances
    3 - impedances
    4 - cycles
    5 - origin segments - for any to_idx, the origin segment of the shortest path
    6 - last segments - for any to_idx, the last segment of the shortest path
    '''
    # prepare the arrays
    n = len(node_edge_map)
    tree_map = np.full((n, 7), np.nan, dtype=np.float32)
    tree_map[:, 0] = 0
    tree_map[:, 2] = np.inf
    tree_map[:, 3] = np.inf
    tree_map[:, 4] = 0
    # prepare proxies
    tree_nodes = tree_map[:, 0]
    tree_preds = tree_map[:, 1]
    tree_dists = tree_map[:, 2]
    tree_imps = tree_map[:, 3]
    tree_cycles = tree_map[:, 4]
    tree_origin_seg = tree_map[:, 5]
    tree_last_seg = tree_map[:, 6]
    # when doing angular, need to keep track of out bearings
    out_bearings = np.full(n, np.nan, dtype=np.float32)
    # keep track of visited edges
    tree_edges = np.full(len(edge_data), False)
    # the starting node's impedance and distance will be zero
    tree_imps[src_idx] = 0
    tree_dists[src_idx] = 0
    # prep the active list and add the source index
    active = List.empty_list(types.int64)
    active.append(src_idx)
    # this loops continues until all nodes within the max distance have been discovered and processed
    while len(active):
        # iterate the currently active indices and find the one with the smallest distance
        min_nd_idx = None
        min_imp = np.inf
        for idx, nd_idx in enumerate(active):
            imp = tree_imps[nd_idx]
            if imp < min_imp:
                min_imp = imp
                min_nd_idx = nd_idx
        # needs this step with explicit cast to int for numpy type inference
        active_nd_idx = int(min_nd_idx)
        # the currently processed node can now be removed from the active list and added to the processed list
        active.remove(active_nd_idx)
        # add to processed nodes
        tree_nodes[active_nd_idx] = True
        # iterate the node's neighbours
        for edge_idx in node_edge_map[active_nd_idx]:
            # get the edge's properties
            start_nd, end_nd, seg_len, seg_ang, seg_imp_fact, seg_in_bear, seg_out_bear = edge_data[edge_idx]
            # cast to int for indexing
            nb_nd_idx = int(end_nd)
            # don't follow self-loops
            if nb_nd_idx == active_nd_idx:
                # add edge to active (used for segment methods)
                tree_edges[edge_idx] = True
                continue
            # don't visit predecessor nodes - otherwise successive nodes revisit out-edges to previous (neighbour) nodes
            if nb_nd_idx == tree_preds[active_nd_idx]:
                continue
            # only add edge to active if the neighbour node has not been processed previously
            # i.e. single direction only
            # i.e. if neighbour node has been processed all out edges have already been explored
            if not tree_nodes[nb_nd_idx]:
                tree_edges[edge_idx] = True
            if not angular:
                # if edge has not been claimed AND the neighbouring node has already been discovered, then it is a cycle
                # do before distance cutoff because this node and the neighbour can respectively be within max distance
                # even if cumulative distance across this edge (via non-shortest path) exceeds distance
                # in some cases all distances are run at once, so keep behaviour consistent by
                # designating the farthest node (but via the shortest distance) as the cycle node
                if not np.isnan(tree_preds[nb_nd_idx]):
                    # bump farther location - prevents mismatching if cycle exceeds threshold in one direction or another
                    if tree_dists[active_nd_idx] <= tree_dists[nb_nd_idx]:
                        tree_cycles[nb_nd_idx] += 0.5
                    else:
                        tree_cycles[active_nd_idx] += 0.5
            # impedance and distance is previous plus new
            if not angular:
                impedance = tree_imps[active_nd_idx] + seg_len * seg_imp_fact
            else:
                # angular impedance include two parts:
                # A - turn from prior simplest-path route segment
                # B - angular change across current segment
                if active_nd_idx == src_idx:
                    turn = 0
                else:
                    turn = np.abs((seg_in_bear - out_bearings[active_nd_idx] + 180) % 360 - 180)
                impedance = tree_imps[active_nd_idx] + (turn + seg_ang) * seg_imp_fact
            dist = tree_dists[active_nd_idx] + seg_len
            # add the neighbour to active if undiscovered but only if less than max threshold
            if np.isnan(tree_preds[nb_nd_idx]) and dist <= max_dist:
                active.append(nb_nd_idx)
            # if impedance less than prior, update
            # this will also happen for the first nodes that overshoot the boundary
            # they will not be explored further because they have not been added to active
            if impedance < tree_imps[nb_nd_idx]:
                tree_imps[nb_nd_idx] = impedance
                tree_dists[nb_nd_idx] = dist
                tree_preds[nb_nd_idx] = active_nd_idx
                out_bearings[nb_nd_idx] = seg_out_bear
                # chain through origin segs - identifies which segment a particular shortest path originated from
                if active_nd_idx == src_idx:
                    tree_origin_seg[nb_nd_idx] = edge_idx
                else:
                    tree_origin_seg[nb_nd_idx] = tree_origin_seg[active_nd_idx]
                # keep track of last seg
                tree_last_seg[nb_nd_idx] = edge_idx
    # the returned active edges contain activated edges, but only in direction of shortest-path discovery
    return tree_map, tree_edges


@njit(cache=True, fastmath=True)
def _find_edge_idx(node_edge_map: Dict, edge_data: np.ndarray, start_nd_idx: int, end_nd_idx: int) -> int:
    '''
    Finds an edge from start and end nodes
    '''
    # iterate the start node's edges
    for edge_idx in node_edge_map[start_nd_idx]:
        # check whether the edge's out node matches the target node
        if edge_data[edge_idx, 1] == end_nd_idx:
            return edge_idx


node_close_func_proto = types.FunctionType(types.float64(types.float64, types.float64, types.float64, types.float64))


# node density
@njit(cache=True, fastmath=True)
def node_density(to_dist, to_imp, beta, cycles):
    return 1.0  # return float explicitly


# node farness
@njit(cache=True, fastmath=True)
def node_farness(to_dist, to_imp, beta, cycles):
    return to_dist


# node cycles
@njit(cache=True, fastmath=True)
def node_cycles(to_dist, to_imp, beta, cycles):
    return cycles


# node harmonic
@njit(cache=True, fastmath=True)
def node_harmonic(to_dist, to_imp, beta, cycles):
    return 1.0 / to_imp


# node beta weighted
@njit(cache=True, fastmath=True)
def node_beta(to_dist, to_imp, beta, cycles):
    return np.exp(beta * to_dist)


# node harmonic angular
@njit(cache=True, fastmath=True)
def node_harmonic_angular(to_dist, to_imp, beta, cycles):
    a = 1 + (to_imp / 180)
    return 1.0 / a


node_betw_func_proto = types.FunctionType(types.float64(types.float64, types.float64))


# node betweenness
@njit(cache=True, fastmath=True)
def node_betweenness(to_dist, beta):
    return 1.0  # return float explicitly


# node betweenness beta weighted
@njit(cache=True, fastmath=True)
def node_betweenness_beta(to_dist, beta):
    '''
    distance is based on distance between from and to vertices
    thus potential spatial impedance via between vertex
    '''
    return np.exp(beta * to_dist)


'''
NODE MAP:
0 - x
1 - y
2 - live
EDGE MAP:
0 - start node
1 - end node
2 - length in metres
3 - sum of angular travel along length
4 - impedance factor
5 - in bearing
6 - out bearing
'''


@njit(cache=False, fastmath=True)
def local_node_centrality(node_data: np.ndarray,
                          edge_data: np.ndarray,
                          node_edge_map: Dict,
                          distances: np.ndarray,
                          betas: np.ndarray,
                          measure_keys: tuple,
                          angular: bool = False,
                          suppress_progress: bool = False) -> np.ndarray:
    # integrity checks
    checks.check_distances_and_betas(distances, betas)
    checks.check_network_maps(node_data, edge_data, node_edge_map)
    # gather functions
    close_funcs = List.empty_list(node_close_func_proto)
    close_idxs = []
    betw_funcs = List.empty_list(node_betw_func_proto)
    betw_idxs = []
    for m_idx, m_key in enumerate(measure_keys):
        if not angular:
            # closeness keys
            if m_key == 'node_density':
                close_funcs.append(node_density)
                close_idxs.append(m_idx)
            elif m_key == 'node_farness':
                close_funcs.append(node_farness)
                close_idxs.append(m_idx)
            elif m_key == 'node_cycles':
                close_funcs.append(node_cycles)
                close_idxs.append(m_idx)
            elif m_key == 'node_harmonic':
                close_funcs.append(node_harmonic)
                close_idxs.append(m_idx)
            elif m_key == 'node_beta':
                close_funcs.append(node_beta)
                close_idxs.append(m_idx)
            # betweenness keys
            elif m_key == 'node_betweenness':
                betw_funcs.append(node_betweenness)
                betw_idxs.append(m_idx)
            elif m_key == 'node_betweenness_beta':
                betw_funcs.append(node_betweenness_beta)
                betw_idxs.append(m_idx)
            else:
                raise ValueError('''
                Unable to match requested centrality measure key against available options.
                Shortest-path measures can't be mixed with simplest-path measures.
                Set angular=True if using simplest-path measures.''')
        else:
            # aggregative keys
            if m_key == 'node_harmonic_angular':
                close_funcs.append(node_harmonic_angular)
                close_idxs.append(m_idx)
            # betweenness keys
            elif m_key == 'node_betweenness_angular':
                betw_funcs.append(node_betweenness)
                betw_idxs.append(m_idx)
            else:
                raise ValueError('''
                Unable to match requested centrality measure key against available options.
                Shortest-path measures can't be mixed with simplest-path measures.
                Set angular=False if using shortest-path measures.''')
    # prepare variables
    n = len(node_data)
    d_n = len(distances)
    k_n = len(measure_keys)
    measures_data = np.full((k_n, d_n, n), 0.0, dtype=np.float32)
    global_max_dist = float(np.nanmax(distances))
    nodes_live = node_data[:, 2]
    # progress steps
    steps = int(n / 10000)
    # iterate through each vert and calculate the shortest path tree
    for src_idx in range(n):
        # numba no object mode can only handle basic printing
        # note that progress bar adds a performance penalty
        if not suppress_progress:
            checks.progress_bar(src_idx, n, steps)
        # only compute for live nodes
        if not nodes_live[src_idx]:
            continue
        '''
        Shortest tree dijkstra        
        Predecessor map is based on impedance heuristic - which can be different from metres
        Distance map in metres still necessary for defining max distances and computing equivalent distance measures
        RETURNS A SHORTEST PATH TREE MAP:
        0 - processed nodes
        1 - predecessors
        2 - distances
        3 - impedances
        4 - cycles
        5 - origin segments
        6 - last segments
        '''
        tree_map, tree_edges = shortest_path_tree(edge_data,
                                                  node_edge_map,
                                                  src_idx,
                                                  max_dist=global_max_dist,
                                                  angular=angular)
        tree_nodes = np.where(tree_map[:, 0])[0]
        tree_preds = tree_map[:, 1]
        tree_dists = tree_map[:, 2]
        tree_imps = tree_map[:, 3]
        tree_cycles = tree_map[:, 4]
        # process each reachable node
        for to_idx in tree_nodes:
            # skip self node
            if to_idx == src_idx:
                continue
            # unpack impedance and distance for to index
            to_imp = tree_imps[to_idx]
            to_dist = tree_dists[to_idx]
            cycles = tree_cycles[to_idx]
            # do not proceed if no route available
            if np.isinf(to_dist):
                continue
            # calculate closeness centralities
            if close_funcs:
                for d_idx in range(len(distances)):
                    dist_cutoff = distances[d_idx]
                    beta = betas[d_idx]
                    if to_dist <= dist_cutoff:
                        for m_idx, close_func in zip(close_idxs, close_funcs):
                            measures_data[m_idx, d_idx, src_idx] += close_func(to_dist, to_imp, beta, cycles)
            # only process in one direction
            if to_idx < src_idx:
                continue
            # calculate betweenness centralities
            if betw_funcs:
                # only counting truly between vertices, not starting and ending verts
                inter_idx = int(tree_preds[to_idx])
                while True:
                    # break out of while loop if the intermediary has reached the source node
                    if inter_idx == src_idx:
                        break
                    # iterate the distance thresholds
                    for d_idx in range(len(distances)):
                        dist_cutoff = distances[d_idx]
                        beta = betas[d_idx]
                        # check threshold
                        if tree_dists[to_idx] <= dist_cutoff:
                            # iterate betweenness functions
                            for m_idx, betw_func in zip(betw_idxs, betw_funcs):
                                measures_data[m_idx, d_idx, inter_idx] += betw_func(to_dist, beta)
                    # follow the chain
                    inter_idx = int(tree_preds[inter_idx])
    return measures_data


segment_func_proto = types.FunctionType(types.float64(types.float64,
                                                      types.float64,
                                                      types.float64,
                                                      types.float64,
                                                      types.float64))


# segment density
@njit(cache=True, fastmath=True)
def segment_density(n, m, n_imp, m_imp, beta):
    return m - n


# segment harmonic
@njit(cache=True, fastmath=True)
def segment_harmonic(n, m, n_imp, m_imp, beta):
    if n_imp < 1:
        return np.log(m_imp)
    else:
        return np.log(m_imp) - np.log(n_imp)


# segment beta
@njit(cache=True, fastmath=True)
def segment_beta(n, m, n_imp, m_imp, beta):
    if beta == -0.0:
        return m_imp - n_imp
    else:
        return (np.exp(beta * m_imp) - np.exp(beta * n_imp)) / beta


@njit(cache=False, fastmath=True)
def local_segment_centrality(node_data: np.ndarray,
                             edge_data: np.ndarray,
                             node_edge_map: Dict,
                             distances: np.ndarray,
                             betas: np.ndarray,
                             measure_keys: tuple,
                             angular: bool = False,
                             suppress_progress: bool = False) -> np.ndarray:
    # integrity checks
    checks.check_distances_and_betas(distances, betas)
    checks.check_network_maps(node_data, edge_data, node_edge_map)
    # gather functions
    close_funcs = List.empty_list(segment_func_proto)
    close_idxs = []
    betw_idxs = []
    for m_idx, m_key in enumerate(measure_keys):
        if not angular:
            # segment keys
            if m_key == 'segment_density':
                close_funcs.append(segment_density)
                close_idxs.append(m_idx)
            elif m_key == 'segment_harmonic':
                close_funcs.append(segment_harmonic)
                close_idxs.append(m_idx)
            elif m_key == 'segment_beta':
                close_funcs.append(segment_beta)
                close_idxs.append(m_idx)
            elif m_key == 'segment_betweenness':
                # only one version of shortest path betweenness - no need for func
                betw_idxs.append(m_idx)
            else:
                raise ValueError('''
                    Unable to match requested centrality measure key against available options.
                    Shortest-path measures can't be mixed with simplest-path measures.
                    Set angular=True if using simplest-path measures. 
                ''')
        else:
            # segment keys
            if m_key == 'segment_harmonic_hybrid':
                # only one version of simplest path closeness - no need for func
                close_idxs.append(m_idx)
            elif m_key == 'segment_betweeness_hybrid':
                # only one version of simplest path betweenness - no need for func
                betw_idxs.append(m_idx)
            else:
                raise ValueError('''
                    Unable to match requested centrality measure key against available options.
                    Shortest-path measures can't be mixed with simplest-path measures.
                    Set angular=False if using shortest-path measures. 
                ''')
    # prepare variables
    n = len(node_data)
    d_n = len(distances)
    k_n = len(measure_keys)
    measures_data = np.full((k_n, d_n, n), 0.0, dtype=np.float32)
    global_max_dist = float(np.nanmax(distances))
    nodes_live = node_data[:, 2]
    # progress steps
    steps = int(n / 10000)
    # iterate through each vert and calculate the shortest path tree
    for src_idx in range(n):
        # numba no object mode can only handle basic printing
        # note that progress bar adds a performance penalty
        if not suppress_progress:
            checks.progress_bar(src_idx, n, steps)
        # only compute for live nodes
        if not nodes_live[src_idx]:
            continue
        '''
        Shortest tree dijkstra        
        Predecessor map is based on impedance heuristic - which can be different from metres
        Distance map in metres still necessary for defining max distances and computing equivalent distance measures
        RETURNS A SHORTEST PATH TREE MAP:
        0 - processed nodes
        1 - predecessors
        2 - distances
        3 - impedances
        4 - cycles
        5 - origin segments
        6 - last segments
        '''
        tree_map, tree_edges = shortest_path_tree(edge_data,
                                                  node_edge_map,
                                                  src_idx,
                                                  max_dist=global_max_dist,
                                                  angular=angular)
        tree_nodes = np.where(tree_map[:, 0])[0]
        tree_preds = tree_map[:, 1]
        tree_dists = tree_map[:, 2]
        tree_imps = tree_map[:, 3]
        tree_origin_seg = tree_map[:, 5]
        tree_last_seg = tree_map[:, 6]
        '''
        can't do edge processing as part of shortest tree because all shortest paths have to be resolved first
        hence visiting all processed edges and extrapolating information
        NOTES:
        1. the above shortest tree algorithm only tracks edges in one direction - i.e. no duplication
        2. dijkstra sorts all active nodes by distance: explores from near to far: edges discovered accordingly
        '''
        # only build edge data if necessary
        if close_idxs:
            for edge_idx in np.where(tree_edges)[0]:
                # unpack the edge data
                seg_n_nd, seg_m_nd, seg_len, seg_ang, seg_imp_fact, seg_in_bear, seg_out_bear = edge_data[edge_idx]
                n_nd_idx = int(seg_n_nd)
                m_nd_idx = int(seg_m_nd)
                n_imp = tree_imps[n_nd_idx]
                m_imp = tree_imps[m_nd_idx]
                n_dist = tree_dists[n_nd_idx]
                m_dist = tree_dists[m_nd_idx]
                # don't process unreachable segments
                if np.isinf(n_dist) and np.isinf(m_dist):
                    continue
                '''
                shortest path (non-angular) uses a split segment workflow
                the split workflow allows for non-shortest-path edges to be approached from either direction
                i.e. the shortest path to node "b" isn't necessarily via node "a"
                the edge is then split at the farthest point from either direction and apportioned either way
                if the segment is on the shortest path then the second segment will squash down to naught
                '''
                if not angular:
                    '''
                    dijkstra discovers edges from near to far (sorts before popping next node)
                    i.e. this sort may be unnecessary?
                    '''
                    # sort where a < b
                    if n_imp <= m_imp:
                        a = tree_dists[n_nd_idx]
                        a_imp = tree_imps[n_nd_idx]
                        b = tree_dists[m_nd_idx]
                        b_imp = tree_imps[m_nd_idx]
                    else:
                        a = tree_dists[m_nd_idx]
                        a_imp = tree_imps[m_nd_idx]
                        b = tree_dists[n_nd_idx]
                        b_imp = tree_imps[n_nd_idx]
                    # get the max distance along the segment: seg_len = (m - start_len) + (m - end_len)
                    # c and d variables can diverge per beneath
                    c = d = (seg_len + a + b) / 2
                    # c | d impedance should technically be the same if computed from either side
                    c_imp = d_imp = a_imp + (c - a) * seg_imp_fact
                    # iterate the distance and beta thresholds - from large to small for threshold snipping
                    for d_idx in range(len(distances) - 1, -1, -1):
                        dist_cutoff = distances[d_idx]
                        beta = betas[d_idx]
                        '''
                        if c or d are greater than the distance threshold, then the segments are "snipped"
                        '''
                        # a to c segment
                        if a <= dist_cutoff:
                            if c > dist_cutoff:
                                c = dist_cutoff
                                c_imp = a_imp + (dist_cutoff - a) * seg_imp_fact
                            for m_idx, close_func in zip(close_idxs, close_funcs):
                                measures_data[m_idx, d_idx, src_idx] += close_func(a, c, a_imp, c_imp, beta)
                        # a to b segment - if on the shortest path then b == d, in which case, continue
                        if b == d:
                            continue
                        if b <= dist_cutoff:
                            if d > dist_cutoff:
                                d = dist_cutoff
                                d_imp = b_imp + (dist_cutoff - b) * seg_imp_fact
                            for m_idx, close_func in zip(close_idxs, close_funcs):
                                measures_data[m_idx, d_idx, src_idx] += close_func(b, d, b_imp, d_imp, beta)
                else:
                    '''
                    there is a different workflow for angular - uses single segment (no segment splitting)
                    this is because the simplest path onto the entire length of segment is from the lower impedance end
                    this assumes segments are relatively straight, overly complex to subdivide segments for spliting...
                    '''
                    # only a single case existing for angular version so no need for abstracted functions
                    # there are three scenarios:
                    # 1) e is the predecessor for f
                    if n_nd_idx == src_idx or tree_preds[m_nd_idx] == n_nd_idx:
                        e = tree_dists[n_nd_idx]
                        f = tree_dists[m_nd_idx]
                        # if travelling via n, then m = n_imp + seg_ang
                        # calculations are based on segment length / angle
                        # i.e. need to decide whether to base angular change on entry vs exit impedance
                        # else take midpoint of segment as ballpark for average, which is the course taken here
                        # i.e. exit impedance minus half segment impedance
                        ang = m_imp - seg_ang / 2
                    # 2) f is the predecessor for e
                    elif m_nd_idx == src_idx or tree_preds[n_nd_idx] == m_nd_idx:
                        e = tree_dists[m_nd_idx]
                        f = tree_dists[n_nd_idx]
                        ang = n_imp - seg_ang / 2  # per above
                    # 3) neither of the above
                    # get the approach angles for either side and compare to find the least inwards impedance
                    # this involves impedance up to entrypoint either side plus respective turns onto the segment
                    else:
                        # get the out bearing from the predecessor and calculate the turn onto current seg's in bearing
                        # find n's predecessor
                        n_pred_idx = int(tree_preds[n_nd_idx])
                        # find the edge from n's predecessor to n
                        e_i = _find_edge_idx(node_edge_map, edge_data, n_pred_idx, n_nd_idx)
                        # get the predecessor edge's outwards bearing at index 6
                        n_pred_out_bear = edge_data[int(e_i), 6]
                        # calculating the turn into this segment from the predecessor's out bearing
                        n_turn_in = np.abs((seg_in_bear - n_pred_out_bear + 180) % 360 - 180)
                        # then add the turn-in to the aggregated impedance at n
                        # i.e. total angular impedance onto this segment
                        # as above two scenarios, adding half of angular impedance for segment as avg between in / out
                        n_ang = n_imp + n_turn_in + seg_ang / 2
                        # repeat for the other side other side
                        # per original n -> m edge destructuring: m is the node in the outwards bound direction
                        # i.e. need to first find the corresponding edge in the opposite m -> n direction of travel
                        # this gives the correct inwards bearing as if m were the entry point
                        opp_i = _find_edge_idx(node_edge_map, edge_data, m_nd_idx, n_nd_idx)
                        # now that the opposing edge is known, we can fetch the inwards bearing at index 5 (not 6)
                        opp_in_bear = edge_data[int(opp_i), 5]
                        # find m's predecessor
                        m_pred_idx = int(tree_preds[m_nd_idx])
                        # we can now go ahead and find m's predecessor edge
                        e_i = _find_edge_idx(node_edge_map, edge_data, m_pred_idx, m_nd_idx)
                        # get the predecessor edge's outwards bearing at index 6
                        m_pred_out_bear = edge_data[int(e_i), 6]
                        # and calculate the turn-in from m's predecessor onto the m inwards bearing
                        m_turn_in = np.abs((opp_in_bear - m_pred_out_bear + 180) % 360 - 180)
                        # then add to aggregated impedance at m
                        m_ang = m_imp + m_turn_in + seg_ang / 2
                        # the distance and angle are based on the smallest angular impedance onto the segment
                        # select by shortest distance in event angular impedances are identical from either direction
                        if n_ang == m_ang:
                            if n_dist <= m_dist:
                                e = tree_dists[n_nd_idx]
                                ang = n_ang
                            else:
                                e = tree_dists[m_nd_idx]
                                ang = m_ang
                        elif n_ang < m_ang:
                            e = tree_dists[n_nd_idx]
                            ang = n_ang
                        else:
                            e = tree_dists[m_nd_idx]
                            ang = m_ang
                        # f is the entry distance plus segment length
                        f = e + seg_len
                    # iterate the distance thresholds - from large to small for threshold snipping
                    for d_idx in range(len(distances) - 1, -1, -1):
                        dist_cutoff = distances[d_idx]
                        if e <= dist_cutoff:
                            if f > dist_cutoff:
                                f = dist_cutoff
                            # uses segment length as base (in this sense hybrid)
                            # intentionally not using integral because conflates harmonic shortest-path w. simplest
                            # there is only one case for angular - no need to abstract to func
                            for m_idx in close_idxs:
                                # transform - prevents division by zero
                                agg_ang = 1 + (ang / 180)
                                # then aggregate - angular uses distances explicitly
                                measures_data[m_idx, d_idx, src_idx] += (f - e) / agg_ang
        if betw_idxs:
            # prepare a list of neighbouring nodes
            nb_nodes = List.empty_list(types.int64)
            for edge_idx in node_edge_map[src_idx]:
                out_nd_idx = int(edge_data[edge_idx][1])  # to node is index 1
                nb_nodes.append(out_nd_idx)
            # betweenness keys computed per to_idx
            for to_idx in tree_nodes:
                # only process in one direction
                if to_idx < src_idx:
                    continue
                # skip self node
                if to_idx == src_idx:
                    continue
                # skip direct neighbours (no nodes between)
                if to_idx in nb_nodes:
                    continue
                # distance - do not proceed if no route available
                to_dist = tree_dists[to_idx]
                if np.isinf(to_dist):
                    continue
                '''
                BETWEENNESS
                segment versions only agg first and last segments
                the distance decay is based on the distance between the src segment and to segment
                i.e. willingness of people to walk between src and to segments
                
                betweenness is aggregated to intervening nodes based on above distances and decays
                other sections (in between current first and last) are respectively processed from other to nodes
                
                distance thresholds are computed using the innner as opposed to outer edges of the segments
                '''
                o_seg_len = edge_data[int(tree_origin_seg[to_idx])][2]
                l_seg_len = edge_data[int(tree_last_seg[to_idx])][2]
                min_span = to_dist - o_seg_len - l_seg_len
                # calculate traversal distances from opposing segments
                o_1 = min_span
                o_2 = min_span + o_seg_len
                l_1 = min_span
                l_2 = min_span + l_seg_len
                # betweenness - only counting truly between vertices, not starting and ending verts
                inter_idx = int(tree_preds[to_idx])
                while True:
                    # break out of while loop if the intermediary has reached the source node
                    if inter_idx == src_idx:
                        break
                    # iterate the distance thresholds - from large to small for threshold snipping
                    for d_idx in range(len(distances) - 1, -1, -1):
                        dist_cutoff = distances[d_idx]
                        beta = betas[d_idx]
                        if min_span <= dist_cutoff:
                            # prune if necessary
                            if o_2 > dist_cutoff:
                                o_2 = dist_cutoff
                            if l_2 > dist_cutoff:
                                l_2 = dist_cutoff
                            # only one version for betweenness for respective angular / non angular
                            # i.e. no need to abstract to function
                            for m_idx in betw_idxs:
                                if not angular:
                                    # catch division by zero
                                    if beta == -0.0:
                                        auc = o_2 - o_1 + l_2 - l_1
                                    else:
                                        auc = (np.exp(beta * o_2) -
                                               np.exp(beta * o_1)) / beta + \
                                              (np.exp(beta * l_2) -
                                               np.exp(beta * l_1)) / beta
                                    measures_data[m_idx, d_idx, inter_idx] += auc
                                else:
                                    bt_ang = 1 + tree_imps[to_idx] / 180
                                    pt_a = o_2 - o_1
                                    pt_b = l_2 - l_1
                                    measures_data[m_idx, d_idx, inter_idx] += (pt_a + pt_b) / bt_ang
                    # follow the chain
                    inter_idx = int(tree_preds[inter_idx])
    return measures_data



"""
A collection of functions for the generation of mock data. This module is predominately used for writing code tests,
but can otherwise be useful for demonstration and utility purposes.
"""
import logging
import string
from typing import Tuple

import networkx as nx
import numpy as np
import pytest
import requests
from shapely import geometry
import utm

from cityseer.tools import graphs

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def mock_graph(wgs84_coords: bool = False) -> nx.MultiGraph:
    """
    Generates a `NetworkX` `MultiGraph` for testing or experimentation purposes.

    Parameters
    ----------
    wgs84_coords
        If set to `True`, the `x` and `y` attributes will be in [WGS84](https://epsg.io/4326) geographic coordinates
        instead of a projected cartesion coordinate system. By default False

    Returns
    -------
    nx.MultiGraph
        A `NetworkX` `MultiGraph` with `x` and `y` node attributes.
    
    Notes
    -----

    ```python
    from cityseer.tools import mock, plot
    networkX_multigraph = mock.mock_graph()
    plot.plot_nX(networkX_multigraph)
    ```

    ![Example graph](../.vitepress/plots/images/graph_example.png)
    _Mock graph._

    """

    networkX_multigraph = nx.MultiGraph()

    nodes = [
        (0, {'x': 700700, 'y': 5719700}),
        (1, {'x': 700610, 'y': 5719780}),
        (2, {'x': 700460, 'y': 5719700}),
        (3, {'x': 700520, 'y': 5719820}),
        (4, {'x': 700620, 'y': 5719905}),
        (5, {'x': 700260, 'y': 5719700}),
        (6, {'x': 700320, 'y': 5719850}),
        (7, {'x': 700420, 'y': 5719880}),
        (8, {'x': 700460, 'y': 5719980}),
        (9, {'x': 700580, 'y': 5720030}),
        (10, {'x': 700100, 'y': 5719810}),
        (11, {'x': 700280, 'y': 5719980}),
        (12, {'x': 700400, 'y': 5720030}),
        (13, {'x': 700460, 'y': 5720130}),
        (14, {'x': 700190, 'y': 5720050}),
        (15, {'x': 700350, 'y': 5720200}),
        (16, {'x': 700800, 'y': 5719750}),
        (17, {'x': 700800, 'y': 5719920}),
        (18, {'x': 700900, 'y': 5719820}),
        (19, {'x': 700910, 'y': 5719690}),
        (20, {'x': 700905, 'y': 5720080}),
        (21, {'x': 701000, 'y': 5719870}),
        (22, {'x': 701040, 'y': 5719660}),
        (23, {'x': 701050, 'y': 5719760}),
        (24, {'x': 701000, 'y': 5719980}),
        (25, {'x': 701130, 'y': 5719950}),
        (26, {'x': 701130, 'y': 5719805}),
        (27, {'x': 701170, 'y': 5719700}),
        (28, {'x': 701100, 'y': 5720200}),
        (29, {'x': 701240, 'y': 5719990}),
        (30, {'x': 701300, 'y': 5719760}),
        (31, {'x': 700690, 'y': 5719590}),
        (32, {'x': 700570, 'y': 5719530}),
        (33, {'x': 700820, 'y': 5719500}),
        (34, {'x': 700700, 'y': 5719480}),
        (35, {'x': 700490, 'y': 5719440}),
        (36, {'x': 700580, 'y': 5719360}),
        (37, {'x': 700690, 'y': 5719370}),
        (38, {'x': 700920, 'y': 5719330}),
        (39, {'x': 700780, 'y': 5719300}),
        (40, {'x': 700680, 'y': 5719200}),
        (41, {'x': 700560, 'y': 5719280}),
        (42, {'x': 700450, 'y': 5719300}),
        (43, {'x': 700440, 'y': 5719150}),
        (44, {'x': 700650, 'y': 5719080}),
        (45, {'x': 700930, 'y': 5719110}),
        # cul-de-sacs
        (46, {'x': 701015, 'y': 5719535}),
        (47, {'x': 701100, 'y': 5719480}),
        (48, {'x': 700917, 'y': 5719517}),
        # isolated node
        (49, {'x': 700400, 'y': 5719550}),
        # isolated edge
        (50, {'x': 700700, 'y': 5720100}),
        (51, {'x': 700700, 'y': 5719900}),
        # disconnected looping component
        (52, {'x': 700400, 'y': 5719650}),
        (53, {'x': 700500, 'y': 5719550}),
        (54, {'x': 700400, 'y': 5719450}),
        (55, {'x': 700300, 'y': 5719550}),
        # add a parallel edge
        (56, {'x': 701300, 'y': 5719110})
    ]

    networkX_multigraph.add_nodes_from(nodes)

    edges = [
        (0, 1),
        (0, 16),
        (0, 31),
        (1, 2),
        (1, 4),
        (2, 3),
        (2, 5),
        (3, 4),
        (3, 7),
        (4, 9),
        (5, 6),
        (5, 10),
        (6, 7),
        (6, 11),
        (7, 8),
        (8, 9),
        (8, 12),
        (9, 13),
        (10, 14),
        (10, 43),
        (11, 12),
        (11, 14),
        (12, 13),
        (13, 15),
        (14, 15),
        (15, 28),
        (16, 17),
        (16, 19),
        (17, 18),
        (17, 20),
        (18, 19),
        (18, 21),
        (19, 22),
        (20, 24),
        (20, 28),
        (21, 23),
        (21, 24),
        (22, 23),
        (22, 27),
        (23, 26),
        (24, 25),
        (25, 26),
        (25, 29),
        (26, 27),
        (27, 30),
        (28, 29),
        (29, 30),
        (30, 45),
        (31, 32),
        (31, 33),
        (32, 34),
        (32, 35),
        (33, 34),
        (33, 38),
        (34, 37),
        (35, 36),
        (35, 42),
        (36, 37),
        (36, 41),
        (37, 39),
        (38, 39),
        (38, 45),
        (39, 40),
        (40, 41),
        (40, 44),
        (41, 42),
        (42, 43),
        (43, 44),
        (44, 45),
        # cul-de-sacs
        (22, 46),
        (46, 47),
        (46, 48),
        # isolated edge
        (50, 51),
        # disconnected looping component
        (52, 53),
        (53, 54),
        (54, 55),
        (55, 52),
        # parallel edge
        (45, 56),
        (30, 56)
    ]

    networkX_multigraph.add_edges_from(edges)

    if wgs84_coords:
        for n, d in networkX_multigraph.nodes(data=True):
            easting = d['x']
            northing = d['y']
            # be cognisant of parameter and return order
            # returns in lat, lng order
            lat, lng = utm.to_latlon(easting, northing, 30, 'U')
            networkX_multigraph.nodes[n]['x'] = lng
            networkX_multigraph.nodes[n]['y'] = lat

    return networkX_multigraph


@pytest.fixture
def primal_graph() -> nx.MultiGraph:
    """
    Returns
    -------
    nx.MultiGraph
        A primal `NetworkX` `MultiGraph` for `pytest` tests.
    """
    G_primal = mock_graph()
    G_primal = graphs.nX_simple_geoms(G_primal)
    return G_primal


@pytest.fixture
def dual_graph() -> nx.MultiGraph:
    """
    Returns
    -------
    nx.MultiGraph
        A dual `NetworkX` `MultiGraph` for `pytest` tests.
    """
    G_dual = mock_graph()
    G_dual = graphs.nX_simple_geoms(G_dual)
    G_dual = graphs.nX_to_dual(G_dual)
    return G_dual


@pytest.fixture
def diamond_graph() -> nx.MultiGraph:
    """
    Generates a diamond shaped `NetworkX` `MultiGraph` for testing or experimentation purposes. For manual checks of all
    node and segmentised methods.
    
    Returns
    -------
    nx.MultiGraph
        A `NetworkX` `MultiGraph` with `x` and `y` node attributes.
    
    Notes
    -----

    ```python
    #     3
    #    / \
    #   /   \
    #  /  a  \
    # 1-------2
    #  \  |  /
    #   \ |b/ c
    #    \|/
    #     0
    # a = 100m = 2 * 50m
    # b = 86.60254m
    # c = 100m
    # all inner angles = 60º
    ```

    """
    G_diamond = nx.MultiGraph()
    G_diamond.add_nodes_from([
        (0, {'x': 50, 'y': 0}),
        (1, {'x': 0, 'y': 86.60254}),
        (2, {'x': 100, 'y': 86.60254}),
        (3, {'x': 50, 'y': 86.60254 * 2})
    ])
    G_diamond.add_edges_from([(0, 1), (0, 2), (1, 2), (1, 3), (2, 3)])
    G_diamond = graphs.nX_simple_geoms(G_diamond)
    return G_diamond


def get_graph_extents(networkX_multigraph: nx.MultiGraph) -> Tuple[float, float, float, float]:
    """
    Parameters
    ----------
    networkX_multigraph
        A `NetworkX` `MultiGraph` with `x` and `y` node parameters.

    Returns
    -------
    Tuple
        A tuple of `min_x`, `min_y`, `max_x`, `max_y` values.
    """
    # get min and maxes for x and y
    min_x = np.inf
    max_x = -np.inf
    min_y = np.inf
    max_y = -np.inf

    for n, d in networkX_multigraph.nodes(data=True):
        if d['x'] < min_x:
            min_x = d['x']
        if d['x'] > max_x:
            max_x = d['x']
        if d['y'] < min_y:
            min_y = d['y']
        if d['y'] > max_y:
            max_y = d['y']

    return min_x, min_y, max_x, max_y


def mock_data_dict(networkX_multigraph: nx.MultiGraph, length: int = 50, random_seed: int = None) -> dict:
    """
    Generates a dictionary containing mock data for testing or experimentation purposes.

    Parameters
    ----------
    networkX_multigraph
        A `NetworkX` graph with `x` and `y` attributes. This is used in order to determine the spatial extents of the
        network. The returned data will be within these extents.
    length
        The number of data elements to return in the dictionary, by default 50.
    random_seed
        An optional random seed, by default None.

    Returns
    -------
    dict
        A dictionary where each entry consists of a `key` representing a distinct data point `uid`, and corresponding
        `x`, `y` and `live` values.
    """
    if random_seed is not None:
        np.random.seed(seed=random_seed)

    min_x, min_y, max_x, max_y = get_graph_extents(networkX_multigraph)

    data_dict = {}

    for i in range(length):
        data_dict[i] = {
            'x': np.random.uniform(min_x, max_x),
            'y': np.random.uniform(min_y, max_y)
        }

    return data_dict


def mock_categorical_data(length: int, num_classes: int = 10, random_seed: int = None) -> np.ndarray:
    """
    Generates a `numpy` array containing mock categorical data for testing or experimentation purposes.

    Parameters
    ----------
    length
        The number of categorical elements to return in the array.
    num_classes
        The maximum number of unique classes to return in the randomly assigned categorical data. The classes are
        randomly generated from a pool of unique class labels of length `num_classes`. The number of returned unique
        classes will be less than or equal to `num_classes`. By default 10
    random_seed
        An optional random seed, by default None

    Returns
    -------
    np.ndarray
        A `numpy` array containing categorical data elements. The number of elements will match the `length` parameter.
        The categorical data will consist of randomly selected letters.

    """
    if random_seed is not None:
        np.random.seed(seed=random_seed)

    random_class_str = string.ascii_lowercase
    if num_classes > len(random_class_str):
        raise ValueError(
            f'The requested {num_classes} classes exceeds the max available categorical classes: {len(random_class_str)}')
    random_class_str = random_class_str[:num_classes]

    d = []
    for i in range(length):
        d.append(random_class_str[np.random.randint(0, len(random_class_str))])

    return np.array(d)


def mock_numerical_data(length: int,
                        min: int = 0,
                        max: int = 100000,
                        num_arrs: int = 1,
                        random_seed: int = None) -> np.ndarray:
    """
    Generates a 2d `numpy` array containing mock numerical data for testing or experimentation purposes.

    Parameters
    ----------
    length
        The number of numerical elements to return in the array.
    min
        The (inclusive) minimum value in the `min`, `max` range of randomly generated integers.
    max
        The (exclusive) maximum value in the `min`, `max` range of randomly generated integers.
    num_arrs
        The number of arrays to nest in the returned 2d array.
    random_seed
        An optional random seed, by default None

    Returns
    -------
    np.ndarray
        A 2d `numpy` array containing numerical data elements. The first dimension corresponds to the number of data
        arrays, and is set with the `num_arrs` parameter. The length of the second dimension will represents the number
        of data elements and will match the `length` parameter. The numerical data will consist of randomly selected
        integers.
    """
    if random_seed is not None:
        np.random.seed(seed=random_seed)

    num_data = []
    for i in range(num_arrs):
        num_data.append(np.random.randint(min, high=max, size=length))
    # return a 2d array
    return np.array(num_data, dtype=float)


def mock_species_data(random_seed: int = None) -> Tuple[np.ndarray, np.ndarray]:
    """
    A generator function returning a series of randomly generated counts and corresponding probabilities for testing
    diversity measures. The data is generated in varying lengths from randomly assigned integers between 1 and 10.
    Matching integers are then collapsed into species "classes" with probabilities computed accordingly.

    Parameters
    ----------
    random_seed
        An optional random seed, by default None

    Yields
    -------
    counts
        The number of members for each species class.
    probs
        The probability of encountering the respective species classes.

    Notes
    -----

    ```python
    from cityseer.tools import mock

    for counts, probs in mock.mock_species_data():
        cs = [c for c in counts]
        print(f'c = {cs}')
        ps = [round(p, 3) for p in probs]
        print(f'p = {ps}')

    # c = [1]
    # p = [1.0]

    # c = [1, 1, 2, 2]
    # p = [0.167, 0.167, 0.333, 0.333]

    # c = [3, 2, 1, 1, 1, 3]
    # p = [0.273, 0.182, 0.091, 0.091, 0.091, 0.273]

    # c = [3, 3, 2, 2, 1, 1, 1, 2, 1]
    # p = [0.188, 0.188, 0.125, 0.125, 0.062, 0.062, 0.062, 0.125, 0.062]

    # etc.
    ```
    """
    if random_seed is not None:
        np.random.seed(seed=random_seed)

    for n in range(1, 50, 5):
        data = np.random.randint(1, 10, n)
        unique = np.unique(data)
        counts = np.zeros_like(unique)
        for i, u in enumerate(unique):
            counts[i] = (data == u).sum()
        probs = counts / len(data)

        yield counts, probs


def fetch_osm_response(geom_osm: str, timeout: int = 30, max_tries: int = 3) -> requests.Response:
    """
    Fetches an OSM response.

    Parameters
    ----------
    geom_osm
        A string representing a polygon for the request extents, formatting according to OSM conventions.
    timeout
        An optional timeout, by default 30s
    max_tries
        The number of attempts to fetch a response before raising, by default 3

    Returns
    -------
    requests.Response
        An OSM API response.
    """

    request = f'''
    /* https://wiki.openstreetmap.org/wiki/Overpass_API/Overpass_QL */
    [out:json][timeout:{timeout}];
    /* build spatial_set from highways based on extent */
    way["highway"]
      ["area"!="yes"]
      ["highway"!~"motorway|motorway_link|bus_guideway|escape|raceway|proposed|abandoned|platform|construction"]
      ["service"!~"parking_aisle"]
      (if:
       /* don't fetch roads that don't have sidewalks */
       (t["sidewalk"] != "none" && t["sidewalk"] != "no")
       /* unless foot or bicycles permitted */
       || t["foot"]!="no"
       || (t["bicycle"]!="no" && t["bicycle"]!="unsuitable")
      )
      ["amenity"!~"charging_station|parking|fuel|motorcycle_parking|parking_entrance|parking_space"]
      ["access"!~"private|customers"]
      ["indoor"!="yes"]
      (poly:"{geom_osm}") -> .spatial_set;
    /* build union_set from spatial_set */
    (
      way.spatial_set["highway"];
      way.spatial_set["foot"~"yes|designated"];
      way.spatial_set["bicycle"~"yes|designated"];
    ) -> .union_set;
    /* filter union_set */
    way.union_set -> .filtered_set;
    /* union filtered_set ways with nodes via recursion */
    (
      .filtered_set;
      >;
    );
    /* return only basic info */
    out skel qt;
    '''
    osm_response = None
    while max_tries:
        osm_response = requests.get('https://overpass-api.de/api/interpreter',
                                    timeout=timeout,
                                    params={'data': request})
        # break if OK response
        if osm_response is not None and osm_response.status_code == 200:
            break
        # otherwise try until max_tries is exhausted
        logger.warning(f'Unsuccessful OSM API request response, trying again...')
        max_tries -= 1

    if osm_response is None or not osm_response.status_code == 200:
        raise('Unsuccessful OSM API request.')

    return osm_response


def make_buffered_osm_graph(lng: float, lat: float, buffer: float) -> nx.MultiGraph:
    """

    Prepares a `networkX` `MultiGraph` from an OSM request for a buffered region around a given `lng` and `lat`
    parameter.

    Parameters
    ----------
    lng
        The longitude argument for the request.
    lat
        The latitude argument for the request.
    buffer
        The buffer distance.

    Returns
    -------
    nx.MultiGraph
        A `networkX` `MultiGraph` with `x` and `y` node attributes that have been converted to UTM.
    """
    # cast the WGS coordinates to UTM prior to buffering
    easting, northing, utm_zone_number, utm_zone_letter = utm.from_latlon(lat, lng)
    # create a point, and then buffer
    pt = geometry.Point(easting, northing)
    poly_utm = pt.buffer(buffer)
    # convert back to WGS
    # the polygon is too big for the OSM server, so have to use convex hull then later prune
    geom = [utm.to_latlon(east, north, utm_zone_number, utm_zone_letter) for east, north in
            poly_utm.convex_hull.exterior.coords]
    poly_wgs = geometry.Polygon(geom)
    # format for OSM query
    geom_osm = str.join(' ', [f'{lat} {lng}' for lat, lng in poly_wgs.exterior.coords])
    # generate the query
    osm_response = fetch_osm_response(geom_osm)
    # build graph
    G_wgs = graphs.nX_from_osm(osm_json=osm_response.text)
    # cast to UTM
    G_utm = graphs.nX_wgs_to_utm(G_wgs)

    return G_utm
